from neuron_morphology.constants import *
from neuron_morphology.validation import radius_validation as rv
from neuron_morphology.validation.result import InvalidMorphology
from tests.objects import test_node, test_tree
from tests.validation.validation_test_case import ValidationTestCase
import unittest
from mock import patch


class TestRadiusValidationFunctions(ValidationTestCase):

    """ Tests the functions in radius_validation.py """

    def test_valid_radius_for_soma(self):
        for valid_radius in [35, 36, 100, 999]:
            nodes = [test_node(type=SOMA, radius=valid_radius)]
            errors = rv.validate_radius_threshold(test_tree(nodes, strict_validation=False))
            self.assertEqual(len(errors), 0)

    def test_invalid_radius_for_soma(self):
        for invalid_radius in [-10, 0, 1, 34, 34.999]:
            nodes = [test_node(type=SOMA, radius=invalid_radius)]
            errors = rv.validate_radius_threshold(test_tree(nodes, strict_validation=False))
            self.assertNodeErrors(errors, "The radius must be", [[1]])

    def test_any_radius_valid_for_axon(self):
        for valid_radius in [-10, 0, 3, float('inf')]:
            nodes = [test_node(id=1, type=SOMA, radius=36, parent_node_id=-1),
                     test_node(id=2, type=AXON, radius=valid_radius, parent_node_id=1)]
            errors = rv.validate_radius_threshold(test_tree(nodes, strict_validation=False))
            self.assertEqual(len(errors), 0)

    def test_valid_radius_for_dendrite(self):
        for valid_radius in [-1, 0, 10, 16, 19.999]:
            for dendrite in [BASAL_DENDRITE, APICAL_DENDRITE]:
                nodes = [test_node(id=1, type=SOMA, radius=36, parent_node_id=-1),
                         test_node(id=2, type=dendrite, radius=valid_radius, parent_node_id=1)]
                errors = rv.validate_radius_threshold(test_tree(nodes, strict_validation=False))
                self.assertEqual(len(errors), 0)

    def test_invalid_radius_for_dendrite(self):
        for valid_radius in [20.002, 35, 40, 100, 1000]:
            for dendrite in [BASAL_DENDRITE, APICAL_DENDRITE]:
                nodes = [test_node(id=1, type=SOMA, radius=36, parent_node_id=-1),
                         test_node(id=2, type=dendrite, radius=valid_radius, parent_node_id=1)]
                errors = rv.validate_radius_threshold(test_tree(nodes, strict_validation=False))
                self.assertNodeErrors(errors, "The radius must be", [[2]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_valid_radius_multiple_types(self):
        nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                 test_node(id=2, type=AXON, radius=2.0, parent_node_id=-1),
                 test_node(id=3, type=BASAL_DENDRITE, radius=3.0, parent_node_id=1),
                 test_node(id=4, type=BASAL_DENDRITE, radius=2.0, parent_node_id=3),
                 test_node(id=5, type=APICAL_DENDRITE, radius=3.0, parent_node_id=1),
                 test_node(id=6, type=APICAL_DENDRITE, radius=2.0, parent_node_id=5)]
        test_tree(nodes)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_invalid_radius_multiple_types(self):
        try:
            nodes = [test_node(id=1, type=SOMA, radius=2.0, parent_node_id=-1),
                     test_node(id=2, type=AXON, radius=2.0, parent_node_id=-1),
                     test_node(id=3, type=BASAL_DENDRITE, radius=32.0, parent_node_id=1),
                     test_node(id=4, type=APICAL_DENDRITE, radius=32.0, parent_node_id=1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "The radius must be", [[1], [3], [4]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_absence_of_constriction_for_dendrite(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=12.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=11.0, parent_node_id=2),
                     test_node(id=4, type=dendrite_type, radius=10.0, parent_node_id=3)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_absence_of_constriction_for_dendrite_one_child_less_than_limit(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=10.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=5.0, parent_node_id=2),
                     test_node(id=4, type=dendrite_type, radius=6.0, parent_node_id=3)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_absence_of_constriction_for_dendrite_multiple_children_less_than_limit(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=12.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=11.0, parent_node_id=2),
                     test_node(id=4, type=dendrite_type, radius=10.0, parent_node_id=2)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_existence_of_constriction_for_dendrite_one_child_less_than_limit(self):
        try:
            for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
                nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                         test_node(id=2, type=dendrite_type, radius=1.0, parent_node_id=1)]

                test_tree(nodes, strict_validation=True)

            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Constriction: The radius of types 3 and 4 should not be less "
                                                       "than 2.0px", [[2]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_existence_of_constriction_for_dendrite_multiple_children(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            try:
                nodes = [test_node(id=1, type=SOMA, radius=35.1, parent_node_id=-1),
                         test_node(id=2, type=dendrite_type, radius=2.0, parent_node_id=1),
                         test_node(id=3, type=dendrite_type, radius=2.0, parent_node_id=2),
                         test_node(id=4, type=dendrite_type, radius=1.5, parent_node_id=3),
                         test_node(id=5, type=dendrite_type, radius=1.5, parent_node_id=3)]

                test_tree(nodes, strict_validation=True)

                self.fail("Morphology should have been rejected.")
            except InvalidMorphology as e:
                self.assertNodeErrors(e.validation_errors, "Constriction: The radius of types 3 and 4 should not be "
                                                           "less than 2.0px", [[5], [4]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_absence_of_constriction_for_dendrite_after_node_ten_from_soma(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=10.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=11.0, parent_node_id=2),
                     test_node(id=4, type=dendrite_type, radius=12.0, parent_node_id=3),
                     test_node(id=5, type=dendrite_type, radius=12.0, parent_node_id=4),
                     test_node(id=6, type=dendrite_type, radius=12.0, parent_node_id=5),
                     test_node(id=7, type=dendrite_type, radius=12.0, parent_node_id=6),
                     test_node(id=8, type=dendrite_type, radius=12.0, parent_node_id=7),
                     test_node(id=9, type=dendrite_type, radius=12.0, parent_node_id=8),
                     test_node(id=10, type=dendrite_type, radius=12.0, parent_node_id=9),
                     test_node(id=11, type=dendrite_type, radius=12.0, parent_node_id=10),
                     test_node(id=12, type=dendrite_type, radius=1.5, parent_node_id=11)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_extreme_taper_less_than_eight_nodes_in_segment(self):
        nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                 test_node(id=2, type=BASAL_DENDRITE, radius=2.0, parent_node_id=1),
                 test_node(id=3, type=BASAL_DENDRITE, radius=2.0, parent_node_id=2),
                 test_node(id=4, type=BASAL_DENDRITE, radius=2.0, parent_node_id=3)]

        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_absence_of_extreme_taper_for_dendrite_more_than_eight_nodes_in_one_segment(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=4.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=4.0, parent_node_id=2),
                     test_node(id=4, type=dendrite_type, radius=4.0, parent_node_id=3),
                     test_node(id=5, type=dendrite_type, radius=4.0, parent_node_id=4),
                     test_node(id=6, type=dendrite_type, radius=4.0, parent_node_id=5),
                     test_node(id=7, type=dendrite_type, radius=4.0, parent_node_id=6),
                     test_node(id=8, type=dendrite_type, radius=3.0, parent_node_id=7),
                     test_node(id=9, type=dendrite_type, radius=3.0, parent_node_id=8)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_existence_of_extreme_taper_for_dendrite_more_than_eight_nodes_in_one_segment(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            try:
                nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                         test_node(id=2, type=dendrite_type, radius=12.0, parent_node_id=1),
                         test_node(id=3, type=dendrite_type, radius=12.0, parent_node_id=2),
                         test_node(id=4, type=dendrite_type, radius=12.0, parent_node_id=3),
                         test_node(id=5, type=dendrite_type, radius=12.0, parent_node_id=4),
                         test_node(id=6, type=dendrite_type, radius=12.0, parent_node_id=5),
                         test_node(id=7, type=dendrite_type, radius=12.0, parent_node_id=6),
                         test_node(id=8, type=dendrite_type, radius=2.0, parent_node_id=7),
                         test_node(id=9, type=dendrite_type, radius=2.0, parent_node_id=8)]

                test_tree(nodes, strict_validation=True)
                self.fail("Morphology should have been rejected.")
            except InvalidMorphology as e:
                self.assertNodeErrors(e.validation_errors, "Extreme Taper: For types 3 and 4", [[2, 8]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_absence_of_extreme_taper_for_dendrite_more_than_eight_nodes_multiple_segments(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=4.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=4.0, parent_node_id=2),
                     test_node(id=4, type=dendrite_type, radius=4.0, parent_node_id=3),
                     test_node(id=5, type=dendrite_type, radius=4.0, parent_node_id=4),
                     test_node(id=6, type=dendrite_type, radius=4.0, parent_node_id=5),
                     test_node(id=7, type=dendrite_type, radius=4.0, parent_node_id=6),
                     test_node(id=8, type=dendrite_type, radius=3.0, parent_node_id=7),
                     test_node(id=9, type=dendrite_type, radius=3.0, parent_node_id=8),
                     test_node(id=10, type=dendrite_type, radius=4.0, parent_node_id=1),
                     test_node(id=11, type=dendrite_type, radius=4.0, parent_node_id=10),
                     test_node(id=12, type=dendrite_type, radius=4.0, parent_node_id=11),
                     test_node(id=13, type=dendrite_type, radius=4.0, parent_node_id=12),
                     test_node(id=14, type=dendrite_type, radius=4.0, parent_node_id=13),
                     test_node(id=15, type=dendrite_type, radius=4.0, parent_node_id=14),
                     test_node(id=16, type=dendrite_type, radius=2.0, parent_node_id=15),
                     test_node(id=17, type=dendrite_type, radius=2.0, parent_node_id=16)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_existence_of_extreme_taper_for_dendrite_more_than_eight_nodes_multiple_segments(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            try:
                nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                         test_node(id=2, type=dendrite_type, radius=10.0, parent_node_id=1),
                         test_node(id=3, type=dendrite_type, radius=10.0, parent_node_id=2),
                         test_node(id=4, type=dendrite_type, radius=10.0, parent_node_id=3),
                         test_node(id=5, type=dendrite_type, radius=10.0, parent_node_id=4),
                         test_node(id=6, type=dendrite_type, radius=10.0, parent_node_id=5),
                         test_node(id=7, type=dendrite_type, radius=10.0, parent_node_id=6),
                         test_node(id=8, type=dendrite_type, radius=2.0, parent_node_id=7),
                         test_node(id=9, type=dendrite_type, radius=2.0, parent_node_id=8),
                         test_node(id=10, type=dendrite_type, radius=10.0, parent_node_id=1),
                         test_node(id=11, type=dendrite_type, radius=10.0, parent_node_id=10),
                         test_node(id=12, type=dendrite_type, radius=10.0, parent_node_id=11),
                         test_node(id=13, type=dendrite_type, radius=10.0, parent_node_id=12),
                         test_node(id=14, type=dendrite_type, radius=10.0, parent_node_id=13),
                         test_node(id=15, type=dendrite_type, radius=10.0, parent_node_id=14),
                         test_node(id=16, type=dendrite_type, radius=2.0, parent_node_id=15),
                         test_node(id=17, type=dendrite_type, radius=2.0, parent_node_id=16)]

                test_tree(nodes, strict_validation=True)
                self.fail("Morphology should have been rejected.")
            except InvalidMorphology as e:
                self.assertNodeErrors(e.validation_errors, "Extreme Taper: For types 3 and 4", [[2, 8], [10, 16]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_decreasing_radius_when_going_away_from_soma_dendrite_valid(self):
        nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                 test_node(id=2, type=BASAL_DENDRITE, radius=4.0, parent_node_id=1),
                 test_node(id=3, type=BASAL_DENDRITE, radius=2.0, parent_node_id=2),
                 test_node(id=4, type=APICAL_DENDRITE, radius=4.0, parent_node_id=1),
                 test_node(id=5, type=APICAL_DENDRITE, radius=2.0, parent_node_id=4)]

        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_decreasing_radius_when_going_away_from_soma_dendrite_one_branch_valid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=dendrite_type, radius=2.0, parent_node_id=1),
                     test_node(id=3, type=dendrite_type, radius=4.0, parent_node_id=2)]

            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_decreasing_radius_when_going_away_from_soma_dendrite_multiple_branches_invalid(self):
        for dendrite_type in [BASAL_DENDRITE, APICAL_DENDRITE]:
            try:

                nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                         test_node(id=2, type=dendrite_type, radius=2.0, parent_node_id=1),
                         test_node(id=3, type=dendrite_type, radius=2.0, parent_node_id=2),
                         test_node(id=4, type=dendrite_type, radius=4.0, parent_node_id=2),
                         test_node(id=5, type=dendrite_type, radius=2.0, parent_node_id=1),
                         test_node(id=6, type=dendrite_type, radius=2.0, parent_node_id=4)]

                test_tree(nodes, strict_validation=True)
                self.fail("Morphology should have been rejected.")
            except InvalidMorphology as e:
                self.assertNodeErrors(e.validation_errors, "Radius should have a negative slope for the following type: %s"
                                      % dendrite_type, [[]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_decreasing_radius_when_going_away_from_soma_axon_valid(self):
        nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                 test_node(id=2, type=AXON, radius=4.0, parent_node_id=1),
                 test_node(id=3, type=AXON, radius=4.0, parent_node_id=2),
                 test_node(id=4, type=AXON, radius=4.0, parent_node_id=1),
                 test_node(id=5, type=AXON, radius=4.0, parent_node_id=4)]

        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_decreasing_radius_when_going_away_from_soma_basal_dendrite_invalid(self):
        try:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=BASAL_DENDRITE, radius=2.0, parent_node_id=1),
                     test_node(id=3, type=BASAL_DENDRITE, radius=2.0, parent_node_id=2),
                     test_node(id=4, type=BASAL_DENDRITE, radius=2.0, parent_node_id=3),
                     test_node(id=5, type=BASAL_DENDRITE, radius=2.0, parent_node_id=4),
                     test_node(id=6, type=BASAL_DENDRITE, radius=2.0, parent_node_id=5),
                     test_node(id=7, type=BASAL_DENDRITE, radius=2.0, parent_node_id=6),
                     test_node(id=8, type=BASAL_DENDRITE, radius=2.0, parent_node_id=7),
                     test_node(id=9, type=BASAL_DENDRITE, radius=2.0, parent_node_id=7),
                     test_node(id=10, type=BASAL_DENDRITE, radius=19.0, parent_node_id=8),
                     test_node(id=11, type=BASAL_DENDRITE, radius=19.0, parent_node_id=9),
                     test_node(id=12, type=BASAL_DENDRITE, radius=19.0, parent_node_id=10),
                     test_node(id=13, type=BASAL_DENDRITE, radius=19.0, parent_node_id=11),
                     test_node(id=14, type=BASAL_DENDRITE, radius=19.0, parent_node_id=12)]

            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Radius should have a negative slope for the following type: %s"
                                  % BASAL_DENDRITE, [[]])

    @patch("neuron_morphology.validation.swc_validators", [rv])
    def test_decreasing_radius_when_going_away_from_soma_apical_dendrite_invalid(self):
        try:
            nodes = [test_node(id=1, type=SOMA, radius=36.0, parent_node_id=-1),
                     test_node(id=2, type=APICAL_DENDRITE, radius=2.0, parent_node_id=1),
                     test_node(id=3, type=APICAL_DENDRITE, radius=2.0, parent_node_id=2),
                     test_node(id=4, type=APICAL_DENDRITE, radius=2.0, parent_node_id=3),
                     test_node(id=5, type=APICAL_DENDRITE, radius=2.0, parent_node_id=4),
                     test_node(id=6, type=APICAL_DENDRITE, radius=2.0, parent_node_id=5),
                     test_node(id=7, type=APICAL_DENDRITE, radius=2.0, parent_node_id=6),
                     test_node(id=8, type=APICAL_DENDRITE, radius=2.0, parent_node_id=7),
                     test_node(id=9, type=APICAL_DENDRITE, radius=2.0, parent_node_id=7),
                     test_node(id=10, type=APICAL_DENDRITE, radius=19.0, parent_node_id=8),
                     test_node(id=11, type=APICAL_DENDRITE, radius=19.0, parent_node_id=9),
                     test_node(id=12, type=APICAL_DENDRITE, radius=19.0, parent_node_id=10),
                     test_node(id=13, type=APICAL_DENDRITE, radius=19.0, parent_node_id=11),
                     test_node(id=14, type=APICAL_DENDRITE, radius=19.0, parent_node_id=12)]

            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Radius should have a negative slope for the following type: %s"
                                  % APICAL_DENDRITE, [[]])


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestRadiusValidationFunctions)
    unittest.TextTestRunner(verbosity=5).run(suite)



import marshmallow as mm
from argschema.schemas import ArgSchema, DefaultSchema
from argschema.fields import InputFile, List, Float, Int, Nested, Boolean
from neuron_morphology.transforms.affine_transformer._schemas import AffineDictSchema


def validate_input_affine(data):
    has_slice_transform_dict = bool(data.get("slice_transform_dict"))
    has_slice_transform_list = bool(data.get("slice_transform_list"))
    if has_slice_transform_list and has_slice_transform_dict:
        err_msg = (
            "Provide either slice_transform_list or " "slice_transform_dict, not both"
        )
        raise mm.ValidationError(err_msg)
    elif not has_slice_transform_list and not has_slice_transform_dict:
        err_msg = "Provide either slice_transform_list or slice_transform_dict"
        raise mm.ValidationError(err_msg)


class InputParameters(ArgSchema):
    swc_path = InputFile(
        description="path to swc file for soma location", required=True
    )
    marker_path = InputFile(
        description="path to reconstruction marker file", required=True
    )
    slice_image_flip = Boolean(
        description=(
            "indicates whether the image was flipped relative "
            "to the slice (avg_group_label.name = 'Flip Slice Indicator'"
        ),
        required=True,
    )
    ccf_soma_location = List(
        Float, description="Soma location (x,y,z) coordinates in CCF", required=True, cli_as_single_argument=True,
    )
    slice_transform_list = List(
        Float,
        required=False,
        cli_as_single_argument=True,
        description="List defining the transform defining slice cut angle",
    )
    slice_transform_dict = Nested(
        AffineDictSchema,
        description="Dict defining the transform defining the slice cut angle",
        required=False,
    )
    ccf_path = InputFile(
        description="path to common cortical framework streamline file", required=True
    )

    @mm.validates_schema
    def validate_schema_input(self, data, **kwargs):
        validate_input_affine(data)


class OutputParameters(DefaultSchema):
    inputs = Nested(
        InputParameters,
        description="The parameters argued to this executable",
        required=True,
    )
    tilt_correction = Float(
        description="Tilt correction about x axis to align with streamlines (radians)",
        required=True,
    )
    tilt_transform_dict = Nested(
        AffineDictSchema,
        required=False,
        description="Dictionary defining an affine transform",
    )



import unittest

import numpy as np

import neuron_morphology.features.layer.layer_histogram as layer
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.features.layer.reference_layer_depths import \
    ReferenceLayerDepths
from neuron_morphology.features.layer.layered_point_depths import \
    LayeredPointDepths
from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.constants import AXON, APICAL_DENDRITE

            
            
def data():
            
    """
    S > A > A > A
    v
    P > P > P
        v
        P
    """
    morphology = (
        MorphologyBuilder()
            .root()
                .axon()
                    .axon()
                        .axon().up(3)
                .apical_dendrite()
                    .apical_dendrite()
                        .apical_dendrite().up()
                        .apical_dendrite()
            .build()
    )


    reference_depths = {
        "1": ReferenceLayerDepths(0, 100),
        "2": ReferenceLayerDepths(100, 200),
        "wm": ReferenceLayerDepths(200, 300, False)
    }

    # in this example specimen, layer transitions are universally at 
    # 50 250 350
    point_depths = LayeredPointDepths(
        ids=np.arange(8),
        layer_name=["2", "2", "wm", "wm", "2", "1", "1", "1"],
        depth=[200, 230, 260, 290, 60, 40, 30, 20],
        local_layer_pia_side_depth=[50, 50, 250, 250, 50, 0, 0, 0],
        local_layer_wm_side_depth=[250, 250, np.nan, np.nan, 250, 50, 50, 50],
        point_type=[node["type"] for node in morphology.nodes()]
    )

    return morphology, reference_depths, point_depths


class TestLayerHistogramFeatures(unittest.TestCase):
    
    def setUp(self):
        self.morphology, self.reference_depths, self.point_depths = data()

    def test_normalized_depth_histogram_within_layer(self):
        expected = layer.LayerHistogram( # specimen thickness half reference
            counts=np.array([0, 0, 0, 0, 1, 0, 1, 0, 1, 0]),
            bin_edges=np.arange(0, 110, 10)
        )

        depths = self.point_depths.df[self.point_depths.df["layer_name"] == "1"]

        obtained = layer.normalized_depth_histogram_within_layer(
            point_depths=depths.depth,
            local_layer_pia_side_depths=depths.local_layer_pia_side_depth,
            local_layer_wm_side_depths=depths.local_layer_wm_side_depth,
            reference_layer_depths=self.reference_depths["1"],
            bin_size=10
        )

        assert np.allclose(expected.counts, obtained.counts)
        assert np.allclose(expected.bin_edges, obtained.bin_edges)

    def test_normalized_depth_histograms_across_layers(self):

        _data = Data(
            self.morphology, 
            reference_layer_depths=self.reference_depths,
            layered_point_depths=self.point_depths
        )

        obtained = layer.normalized_depth_histogram(
            _data, bin_size=10
        )

        expected = {
            "1": layer.LayerHistogram(
                counts=np.array([0, 0, 0, 0, 1, 0, 1, 0, 1, 0]), 
                bin_edges=np.arange(0, 110, 10)),
            "2": layer.LayerHistogram(
                counts=np.array([1, 0, 0, 0, 0, 0, 0, 1, 0, 1]), 
                bin_edges=np.arange(100, 210, 10)),
            "wm": layer.LayerHistogram(
                counts=np.array([0, 1, 0, 0, 1, 0, 0, 0, 0, 0]), 
                bin_edges=np.arange(200, 310, 10))
        }

        self.assertEqual(set(expected.keys()), set(obtained.keys()))
        for key, obt in obtained.items():
            expt = expected[key]
            print(key, expt.counts, obt.counts)
            assert np.allclose(expt.counts, obt.counts)
            assert np.allclose(expt.bin_edges, obt.bin_edges)


    def test_histogram_earth_movers_distance(self):
        
        obtained = layer.histogram_earth_movers_distance(
            np.array([0, 0, 0, 1, 1, 0, 0, 0]),
            np.array([0, 0, 0, 0, 0, 0, 1, 1])
        )

        self.assertAlmostEqual(obtained.result, 3)
        self.assertEqual(
            obtained.interpretation, 
            layer.EarthMoversDistanceInterpretation.BothPresent
        )

    def test_histogram_earth_movers_distance_one_zero(self):
        obtained = layer.histogram_earth_movers_distance(
            np.array([0, 0, 0, 1, 1, 0, 0, 0]),
            np.array([0, 0, 0, 0, 0, 0, 0, 0])
        )

        self.assertEqual(obtained.result, 2)
        self.assertEqual(
            obtained.interpretation, 
            layer.EarthMoversDistanceInterpretation.OneEmpty
        )

    def test_histogram_earth_movers_distance_both_zero(self):
        obtained = layer.histogram_earth_movers_distance(
            np.array([0, 0, 0, 0, 0, 0, 0, 0]),
            np.array([0, 0, 0, 0, 0, 0, 0, 0])
        )

        self.assertEqual(obtained.result, 0)
        self.assertEqual(
            obtained.interpretation, 
            layer.EarthMoversDistanceInterpretation.BothEmpty
        )

    def test_earth_movers_distance(self):

        _data = Data(
            self.morphology, 
            reference_layer_depths=self.reference_depths,
            layered_point_depths=self.point_depths
        )

        obtained = layer.earth_movers_distance(
            _data, [AXON], [APICAL_DENDRITE]
        )

        self.assertEqual(set(obtained.keys()), {"2"})
        self.assertEqual(obtained["2"].result, 17)



from typing import Union, Any, Dict

from neuron_morphology.morphology import Morphology

class Data:

    def __init__(self, morphology: Morphology, **other_things):
        """ A placeholder for the "general data" objsect that we will pass 
        into the feature extractor. Guaranteed to have a morphology. Might 
        have other things.
        """

        self.morphology: Morphology = morphology
        for name, value in other_things.items():
            setattr(self, name, value)

    def __hash__(self):
        return hash(id(self))

# Using get_morphology, functions can easily accept either a Data or a 
# Morphology. This derived type expresses that union.
MorphologyLike = Union[Data, Morphology]

def get_morphology(data: MorphologyLike):
    """ Decay a Data to a Morphology, leaving Morphologies untouched
    """

    if isinstance(data, Morphology):
        return data
    return data.morphology



import warnings
import random

from neuron_morphology.morphology import Morphology
from neuron_morphology.constants import (
    SOMA, AXON, APICAL_DENDRITE, BASAL_DENDRITE)

class MorphologyBuilder:
    
    @property
    def next_id(self):
        result = self._id
        self._id += 1
        return result
        
    @property
    def parent_id(self):
        return self._parent_queue[-1] if self._parent_queue else -1

    @property
    def active_node_id(self):
        if self._parent_queue:
            return self._parent_queue[-1]
    
    def __init__(self):
        """ A utility for putting together a morphology object using succinct,
        declarative code. See tests/test_morphology_builder.py for example 
        usage.
        """

        self._id = 0
        self._parent_queue = []
        self.nodes = []
        self.rng = random.Random()
        
    def up(self, by=1):
        """ Terminate a branch. Set the active node to the previous active 
        node's ancestor.

        Parameters
        ----------
        by : how far (up the tree) to set the new active node. Default is the 
            parent of the current node (1). 2 would correspond to the 

        """

        for _ in range(by):
            if self._parent_queue:
                self._parent_queue.pop()
            else:
                warnings.warn(
                    f"this {self.__class__.__name__}'s active node has been "
                    "unset. If you wish to add more nodes from this point "
                    "you must use .root() to create a new root"
                )
        return self
        
    def root(self, x=0, y=0, z=0, node_type=SOMA, radius=1):
        """ Add a new root node (parent -1) to this reconstruction. This will 
        be the new active node.
        """

        self._parent_queue = []
        return self._add_node(x, y, z, node_type, radius)
    
    def child(self, x, y, z, node_type, radius=1):
        """ Add a child node to the current active node. This will become the 
        new active node.
        """

        if not self._parent_queue:
            raise ValueError(
                "attempted to create a child node, but the active node is "
                "unset. If you wish to add more nodes from this point "
                "you must use .root() to create a new root"
            )
        return self._add_node(x, y, z, node_type, radius)
        
    def _add_node(self, x, y, z, node_type, radius):
        """ Add a new node to this morphology.

        Parameters
        ----------
        x : x coordinate of this node's position
        y : y coordinate of this node's position
        z : z coordinate of this node's position
        node_type : one of AXON, SOMA, APICAL_DENDRITE, BASAL_DENDRITE
        radius : describe the size of this node

        """

        node_id = self.next_id

        self.nodes.append({
            "x": x if x is not None else self.rng.uniform(-1, 1),
            "y": y if y is not None else self.rng.uniform(-1, 1),
            "z": z if z is not None else self.rng.uniform(-1, 1),
            "radius": radius,
            "type": node_type,
            "id": node_id,
            "parent": self.parent_id
        })
        self._parent_queue.append(node_id)
        return self

    def axon(self, x=None, y=None, z=None, radius=1):
        """ Convenvience for creating an axon node. Will not create a root.
        """
        return self.child(x, y, z, AXON, radius)

    def apical_dendrite(self, x=None, y=None, z=None, radius=1):
        """ Convenvience for creating an apical dendrite node. Will not create 
        a root.
        """
        return self.child(x, y, z, APICAL_DENDRITE, radius)

    def basal_dendrite(self, x=None, y=None, z=None, radius=1):
        """ Convenvience for creating a basal dendrite node. Will not create a 
        root.
        """
        return self.child(x, y, z, BASAL_DENDRITE, radius)
        
    def build(self):
        """ Construct a Morphology object using this builder. This is a non-
        destructive operation. The Morphology will be validated at this stage.
        """

        return Morphology(
            self.nodes, 
            node_id_cb=lambda node: node["id"],
            parent_id_cb=lambda node: node["parent"]
        )



"""A collection of utilities used by snap polygons to manipulate shapely
objects.
"""
from typing import (
    Optional, Dict, Union, Sequence, Callable, Iterable, Iterator
)
import collections
import math
from functools import partial
import sys
import itertools as it

import shapely.ops
from shapely.geometry.base import BaseGeometry
from shapely.geometry.polygon import Polygon
from shapely.geometry import LineString
import numpy as np
from scipy import ndimage
import rasterio
import rasterio.features

from neuron_morphology.snap_polygons.bounding_box import BoundingBox
from neuron_morphology.snap_polygons.types import (
    PolyType, LineType, TransformType, ensure_polygon, ensure_linestring,
    MultiPolygonResolverType, MultiSurfaceResolvertype
)


def select_largest_subpolygon(
        polygons: Union[Polygon, Iterable[Polygon]],
        error_threshold: float
) -> Polygon:
    """Given a collection of polygons, find the largest by area.

    Parameters
    ----------
    polygons : To be filtered
    error_threshold : If the ratio of the largest polygon to the second
        largest does not meet or exceed this value, reject the largest polygon.

    Returns
    -------
    the largest polygon
    """

    if isinstance(polygons, shapely.geometry.Polygon):
        return polygons

    polygons = [
        poly for poly in polygons
        if poly.area >= sys.float_info.epsilon
    ]

    if len(polygons) == 1:
        return polygons[0]
    elif len(polygons) == 0:
        raise ValueError("No argued geometries with nonzero area")

    largest = (-float("inf"), None)
    second = (-float("inf"), None)
    for subpolygon in polygons:
        area = subpolygon.area

        if area < sys.float_info.epsilon:
            continue

        if area > largest[0]:
            second = largest
            largest = (area, subpolygon)
        elif area > second[0]:
            second = (area, subpolygon)

    ratio = largest[0] / second[0]
    if ratio < error_threshold:
        raise ValueError(
            "no definitive largest polygon. "
            f"{largest[0]} / {second[0]} = {ratio} < {error_threshold}"
        )
    return largest[1]


def safe_linemerge(
        linestrings: Union[LineString, Sequence[LineString]]
) -> LineString:
    """Wrapper around shapely.ops.linemerge that no-ops in case a single
    LineString or length-1 collection is argued.
    """

    if isinstance(linestrings, LineString):
        return linestrings

    if len(linestrings) == 1:
        return linestrings[0]
    elif len(linestrings) == 0:
        raise ValueError("Must argue at least one linestring")

    return shapely.ops.linemerge(linestrings)


class Geometries:
    """ A collection of polygons and lines
    """

    @property
    def default_multipolygon_resolver(self):
        """By default, multiple polygons resulting from operations on these
        geometries are resolved by discarding all but the largest
        """
        return partial(
            select_largest_subpolygon,
            error_threshold=-float("inf")
        )

    @property
    def default_multisurface_resolver(self):
        """By default, multiple surfaces arising from operations on these
        geometries are merged back together (failing if this is not possible).
        """
        return safe_linemerge

    def __init__(self):
        self._close_bounds: Optional[BoundingBox] = None

        self.polygons: Dict[str, Polygon] = {}
        self.surfaces: Dict[str, LineString] = {}

    @property
    def close_bounds(self):
        """The smallest bounding box enclosing these geometries.
        """
        if self._close_bounds is None:
            self._close_bounds = BoundingBox(
                np.inf, np.inf, -np.inf, -np.inf
            )
        return self._close_bounds

    @close_bounds.setter
    def close_bounds(self, value: BoundingBox):
        self._close_bounds = value

    def register_polygon(
            self,
            name: str,
            path: PolyType
    ):
        """ Adds a named polygon path to this object. Updates the close
        bounding box.

        Parameters
        ----------
        name : identifier for this polygon
        path : defines the exterior of this (simple) polygon


        """

        polygon = ensure_polygon(path)
        self.close_bounds.update(*polygon.bounds)

        self.polygons[name] = polygon

    def _register_many(
            self,
            objects: Union[
                Dict[str, Union[LineType, PolyType]],
                Sequence[Dict[str, Union[LineType, PolyType]]]
            ],
            method: Callable[[str, Union[LineType, PolyType]], None]
    ):
        """ Utility for registering many polygons or surfaces. See
        register_polygons and register_surfaces for use.
        """

        if isinstance(objects, collections.abc.Sequence):
            for obj in objects:
                method(obj["name"], obj["path"])

        elif isinstance(objects, collections.abc.Mapping):
            for name, path in objects.items():
                method(name, path)

        else:
            raise TypeError(f"did not understand type: {type(objects)}")

    def register_polygons(
            self,
            polygons: Union[
                Dict[str, PolyType],
                Sequence[Dict[str, PolyType]]
            ]
    ):
        """ utility for registering multiple polygons. See register_polygon
        """
        self._register_many(polygons, self.register_polygon)

    def register_surface(
            self,
            name: str,
            path: LineType
    ):
        """ Adds a line (e.g. the pia/wm surfaces) to this object. Updates
        the bounding box.

        Parameters
        ----------
        name : identifier for this surface
        path : defines the surface

        """

        surface = ensure_linestring(path)
        self.close_bounds.update(*surface.bounds)

        self.surfaces[name] = surface

    def register_surfaces(self, surfaces: Dict[str, LineType]):
        """ utility for registering multiple surfaces. See register_surface
        """
        self._register_many(surfaces, self.register_surface)

    def rasterize(
            self,
            box: Optional[BoundingBox] = None,
            polygons: Union[Sequence[str], bool] = True,
            surfaces: Union[Sequence[str], bool] = False
    ) -> Dict[str, np.ndarray]:
        """ Rasterize one or more owned geometries. Produce a mapping from
        object names to masks.

        Parameters
        ----------
        shape : if provided, the output image shape. Otherwise, use the
            rounded close bounding box shape
        polygons : a list of names. Alternatively all (True) or none (False)
        lines : a list of names. Alternatively all (True) or none (False)

        Notes
        -----
        uses rasterio.features.rasterize

        """

        if box is None:
            box = self.close_bounds

        if polygons is True:
            polygons = list(self.polygons.keys())
        elif polygons is False:
            polygons = []

        if surfaces is True:
            surfaces = list(self.surfaces.keys())
        elif surfaces is False:
            surfaces = []

        stack = {}

        for name in polygons:
            stack[name] = rasterize(self.polygons[name], box)

        for name in surfaces:
            stack[name] = rasterize(self.surfaces[name], box)

        return stack

    def transform(
            self,
            transform: TransformType
    ) -> "Geometries":
        """ Apply a transform to each owned geometry. Return a new collection.

        Parameters
        ----------
        transform : A callable which maps (vertical, horizontal) coordinates to
            new (vertical, horizontal) coordinates.

        """

        out = Geometries()

        for name, poly in self.polygons.items():
            out.register_polygon(name, shapely.ops.transform(transform, poly))

        for name, surf in self.surfaces.items():
            out.register_surface(name, shapely.ops.transform(transform, surf))

        return out

    def fill_gaps(
            self,
            working_scale: float = 1.0,
            multipolygon_resolver: Optional[MultiPolygonResolverType] = None
    ) -> "Geometries":
        """Expand this geometries' polygons to fill its bounding box, using
        distance to assign empty space.

        Parameters
        ----------
        working_scale : The filling is carried out in a raster space, with 1
            pixel corresponding to 1 unit in the coordinate system of your
            polygons. You can optionally rescale the polygons before
            rasterizing.
        multipolygon_resolver : This method might obtain multiple output
            polygons for a given input polygon. This callable collapses them
            into a single geometry. The default selects the largest.

        Returns
        -------
        A copy of this geometries object with the entire bounding box having
        been filled.
        """
        multipolygon_resolver = multipolygon_resolver \
            or self.default_multipolygon_resolver

        scale_to_working = make_scale(working_scale)
        working_geometries = self.transform(scale_to_working)

        raster_stack = working_geometries.rasterize()
        clear_overlaps(raster_stack)
        closest, closest_names = closest_from_stack(raster_stack)
        snapped_polygons = get_snapped_polys(
            closest, closest_names, multipolygon_resolver
        )

        result_geometries = Geometries()
        result_geometries.register_polygons(snapped_polygons)

        translation_from_working = make_translation(
            working_geometries.close_bounds.horigin,
            working_geometries.close_bounds.vorigin
        )
        scale_from_working = make_scale(1.0 / working_scale)

        result_geometries = (
            result_geometries
            .transform(translation_from_working)
            .transform(scale_from_working)
        )
        result_geometries.register_surfaces(self.surfaces)
        return result_geometries

    def cut(
            self,
            template: shapely.geometry.Polygon,
            multipolygon_resolver: Optional[MultiPolygonResolverType] = None,
            multisurface_resolver: Optional[MultiSurfaceResolvertype] = None
    ) -> "Geometries":
        """Crop this Geometries' polygons and surfaces onto a provided template.

        Parameters
        ----------
        template : portions of surfaces and polygons outside this shape will be
            removed
        multipolygon_resolver : This callable is applied to the outputs of
            the intersection operation in order to resolve cases where a
            polygon has been cut into multiple components. The default method
            selects the largest by area.
        multisurface_resolver : As multipolygon resolver, for surfaces. The
            default method attempts to merge the surfaces.

        Returns
        -------
        A copy of this Geometries object, with polygons and surfaces cropped
        """
        multipolygon_resolver = multipolygon_resolver \
            or self.default_multipolygon_resolver
        multisurface_resolver = multisurface_resolver \
            or self.default_multisurface_resolver

        result = Geometries()

        for key, polygon in self.polygons.items():
            polygon = polygon.intersection(template)
            polygon = multipolygon_resolver(polygon)
            result.register_polygon(key, polygon)

        for key, surface in self.surfaces.items():
            surface = surface.intersection(template)
            surface = multisurface_resolver(surface)
            result.register_surface(key, surface)

        return result

    def convex_hull(
            self,
            surfaces: bool = True,
            polygons: bool = True
    ) -> Polygon:
        """Find the convex hull of these geometries.

        Parameters
        ----------
        surfaces : if True, include surfaces in the hull
        polygons : if True, include polygons in the hull

        Returns
        -------
        The convex hull of the included geometries
        """
        geometries: Iterator[BaseGeometry] = iter([])
        if surfaces:
            geometries = it.chain(geometries, self.surfaces.values())
        if polygons:
            geometries = it.chain(geometries, self.polygons.values())

        hull = next(geometries).convex_hull
        for geometry in geometries:
            # why the intermediate hull-taking? Some layer polygons have
            # loops at the corners. This breaks the union operation, since
            # they don't have a defined interior/exterior.
            hull = geometry.convex_hull.union(hull)
        return hull.convex_hull

    def to_json(self) -> Dict:
        """ Write contained polygons to a json-serializable format
        """

        return {
            "polygons": [
                {
                    "name": name,
                    "path": np.array(poly.exterior.coords).tolist()
                }
                for name, poly in self.polygons.items()
            ],
            "surfaces": [
                {
                    "name": name,
                    "path": np.array(surf.coords).tolist()
                }
                for name, surf in self.surfaces.items()
            ]
        }


def rasterize(
        geometry: shapely.geometry.base.BaseGeometry,
        box: BoundingBox
) -> np.array:
    """ Rasterize a shapely object to a grid defined by a provided bounding box.

    Parameters
    ----------
    geometry : to be rasterized
    box : defines the window (in the same coordinate space as the geometry)
        into which the geometry will be rasterized

    Returns
    -------
    A mask, where 1 indicates presence and 0 absence

    """

    box = box.round(origin_via=math.floor, extent_via=math.ceil)
    translate = make_translation(-box.horigin, -box.vorigin)
    geometry = shapely.ops.transform(translate, geometry)
    out_shape = (box.height, box.width)

    return rasterio.features.rasterize(
        [(geometry, 1)],
        out_shape=out_shape
    )


def make_translation(horizontal: float, vertical: float) -> TransformType:
    """Utility for building a 2D translation transform

    Parameters
    ----------
    horizontal : translate by this much along the first axis
    vertical : translate by this much along the second axis

    Returns
    -------
    Function which applies the argued translation
    """
    return lambda ht, vt: (ht + horizontal, vt + vertical)


def make_scale(
        scale: float = 1.0
) -> TransformType:
    """ A utility for making a 2D scale transform, suitable for transforming
    bounding boxes and Geometries

    Parameters
    ----------
    scale : isometric scale factor

    Returns
    -------
    A transform function

    """
    return lambda horizontal, vertical: (horizontal * scale, vertical * scale)


def clear_overlaps(stack: Dict[str, np.ndarray]):
    """ Given a stack of masks, remove all inter-mask overlaps inplace

    Parameters
    ----------
    stack : Keys are names, values are masks (of the same shape). 0 indicates
        absence

    """

    overlaps = np.array(list(stack.values())).sum(axis=0) >= 2

    for image in stack.values():
        image[overlaps] = 0


def closest_from_stack(stack: Dict[str, np.ndarray]):
    """ Given a stack of images describing distance from several objects, find
    the closest object to each pixel.

    Parameters
    ----------
    stack : Keys are names, values are ndarrays (of the same shape). Each pixel
        in the values describes the distance from that pixel to the named
        object

    Returns
    -------
    closest : An integer array whose values are the closest object to each
        pixel
    names : A mapping from the integer codes in the "closest" array to names

    """

    distances = []
    names = {}

    for index, (name, mask) in enumerate(stack.items()):
        distances.append(ndimage.distance_transform_edt(1 - mask))
        names[index + 1] = name

    closest = np.squeeze(np.argmin(distances, axis=0)) + 1
    return closest, names


def get_snapped_polys(
        closest: np.ndarray,
        name_lut: Dict[int, str],
        multipolygon_resolver: MultiPolygonResolverType
) -> Dict[str, Polygon]:
    """ Obtains named shapes from a label image.

    Parameters
    ----------
    closest : label integer with integer codes
    name_lut : look up table from integer codes to string names

    Returns
    -------
    mapping from names to polygons describing each labelled region

    """

    polygons = collections.defaultdict(list)
    for obtained, label in rasterio.features.shapes(closest.astype(np.uint16), connectivity=8):
        key = name_lut[label]
        for coords in obtained["coordinates"]:
            polygons[key].append(Polygon(coords))

    for key in list(polygons.keys()):
        polygons[key] = multipolygon_resolver(polygons[key])

    return polygons


def find_vertical_surfaces(
        polygons: Dict[str, Polygon],
        order: Sequence[str],
        pia: Optional[LineString] = None,
        white_matter: Optional[LineString] = None
):
    """ Given a set of polygons describing cortical layer boundaries, find the
    boundaries between each layer.

    Parameters
    ----------
    polygons : named layer polygons
    order : A sequence of names defining the order of the layer polygons from
        pia to white matter
    pia : The upper (from the perspective of cortex) pia surface.
    white_matter : The lower (from the perspective of cortex) white matter
        surface.

    Returns
    -------
    dictionary whose keys are as "{name}_{side}" and whose values are
        linestrings describing these boundaries.

    """

    names = [name for name in order if name in polygons]
    results = {}

    for index, name in enumerate(names):
        current = polygons[name]
        # up side
        if index == 0 and pia is not None:
            results[f"{name}_pia"] = pia
        else:
            above_layers = [polygons[name] for name in names[:index]]
            results[f"{name}_pia"] = shared_faces(current, above_layers)

        # down side
        if index == len(names) - 1 and white_matter is not None:
            results[f"{name}_wm"] = white_matter
        else:
            below_layers = [polygons[name] for name in names[index + 1:]]
            results[f"{name}_wm"] = shared_faces(current, below_layers)

    return results


def shared_faces(poly: Polygon, others: Iterable[Polygon], snap_tolerance=10) -> LineString:
    """ Given a polygon and a set of other polygons that could be adjacent on
    the same side, find and connect that shared face.

    Parameters
    ----------
    poly : Polygon
        Polygon whose boundary with others we want to identify
    others : list
        List of other Polygons

    Returns
    -------
    LineString representing the shared face
    """

    merged_others = shapely.ops.unary_union(others)
    geom_collection = shapely.ops.shared_paths(
        poly.exterior, merged_others.exterior
    )
    _forward, backward = geom_collection.geoms
    faces = shapely.ops.linemerge(backward)

    # check for multiple components
    if faces.geom_type == "MultiLineString":
        flat_faces_list = list(faces.geoms)
        # Ensure everything is contiguous
        for i, (f, f_prev) in enumerate(zip(flat_faces_list[1:], flat_faces_list[:-1])):
            # figure out how they are oriented
            if f_prev.boundary.geoms[0].distance(f) < f_prev.boundary.geoms[1].distance(f):
                prev_ind = 0
            else:
                prev_ind = 1
            if f.boundary.geoms[0].distance(f_prev) < f.boundary.geoms[1].distance(f_prev):
                # closer at start
                flat_faces_list[i + 1] = shapely.geometry.LineString(list(f_prev.boundary.geoms[prev_ind].coords) + list(f.coords))
            else:
                # closer at end
                flat_faces_list[i + 1] = shapely.geometry.LineString(list(f.coords) + list(f_prev.boundary.geoms[prev_ind].coords))

        faces = shapely.ops.linemerge(flat_faces_list)
    coordinates = list(faces.coords)
    shared_line = ensure_linestring(coordinates)
    return shared_line



from typing import Optional, List

import numpy as np

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.features.statistics.coordinates import COORD_TYPE

from neuron_morphology.feature_extractor.marked_feature import marked
from neuron_morphology.feature_extractor.mark import Geometric


def calculate_coordinate_overlap_from_min_max(coordinates: np.ndarray,
                                              minv: float,
                                              maxv: float,
                                              dimension: int = 1):
    """
        Return the % of coordinates that are above the max,
        between, or below the min

        Parameters
        ----------

        coordinates: np.ndarray with x, y, z columns

        minv: min to check against

        maxv: max to check against

        dimension: dimension to compare (0, 1, 2 for x, y, z), default 1 (y)

    """
    n = coordinates.shape[0]
    above = (coordinates[:, dimension] > maxv).sum() / n
    below = (coordinates[:, dimension] < minv).sum() / n
    overlap = 1 - above - below
    return above, overlap, below


def calculate_coordinate_overlap(coordinates_a,
                                 coordinates_b,
                                 dimension: int = 1):
    """
        Return the % of coordinates_a that are above, overlaping, and below
        coordinates_b, and the same for b over a

        Parameters
        ----------

        coordinates_a: 2d array-like with x, y, z cols

        coordinates_b: 2d array-like with x, y, z cols

        dimension: dimension to compare (0, 1, 2 for x, y, z), default 1 (y)

        Returns
        -------

        dict: a_above_b, a_overlap_b, a_below_b, or
              -1's if coordinates_b is empty
    """
    if not coordinates_b:
        a_above_b, a_overlap_b, a_below_b = (-1, -1, -1)

    else:
        coordinates_a = np.asarray(coordinates_a)
        coordinates_b = np.asarray(coordinates_b)

        min_b = coordinates_b[:, dimension].min()
        max_b = coordinates_b[:, dimension].max()

        a_above_b, a_overlap_b, a_below_b = \
            calculate_coordinate_overlap_from_min_max(
                coordinates_a, min_b, max_b)

    overlap_features = {}
    overlap_features['above'] = a_above_b
    overlap_features['overlap'] = a_overlap_b
    overlap_features['below'] = a_below_b

    return overlap_features


@marked(Geometric)
def overlap(data: Data,
            node_types: Optional[List[int]] = None,
            node_types_to_compare: Optional[List[int]] = None,
            coord_type: COORD_TYPE = COORD_TYPE.NODE,
            dimension: int = 1):
    """
        Compares the locations of node_types to node_types_to_compare
        Calculate % of coordinates of node_types that are
        above, overlapping, and below the coordinates of node_types_to_compare

        Example: calculate_overlap(
                    morphology,
                    node_types=[AXON],
                    node_types_to_compare=[APICAL_DENDRITE, BASAL_DENDRITE])
                will return the percentage of AXON nodes that are above,
                overlapping, and below DENDRITE nodes

        Parameters
        ----------
        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)
        node_types_to_compare: a list of node types (see neuron_morphology constants)
        coord_type: Restrict analysis to specific coordinate type
            (see neuron_morphology.features.statistics.coordinates for options)
        dimension: dimension to compare (0, 1, 2 for x, y, z), default 1 (y)

    """
    coords_a = coord_type.get_coordinates(data.morphology, node_types)
    coords_b = coord_type.get_coordinates(data.morphology, node_types_to_compare)

    overlap_features = calculate_coordinate_overlap(coords_a,
                                                    coords_b,
                                                    dimension=1)
    return overlap_features



import unittest

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.marked_feature import (
    specialize
)
from neuron_morphology.feature_extractor.feature_specialization import (
    NEURITE_SPECIALIZATIONS
)
from neuron_morphology.features.intrinsic import (
    num_branches,
    max_branch_order
)


class TestIntrinsic(unittest.TestCase):

    def test_number_of_branches_by_types(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(2, 1, 0)
                                .basal_dendrite(1, 2, 0).up()
                                .basal_dendrite(3, 2, 0).up(2)
                            .axon(2, -1, 0)
                                .axon(1.5, -2, 0).up()
                                .axon(2.5, -2, 0).up(4)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 3, 0).up()
                            .apical_dendrite(1, 3, 0)

                .build()
        )


        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(num_branches, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.num_branches"], 10)
        self.assertEqual(feature_extraction_run.results["axon.num_branches"], 3)
        self.assertEqual(feature_extraction_run.results["dendrite.num_branches"], 6)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.num_branches"], 3)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.num_branches"], 3)


    def test_max_branch_order_by_types(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(2, 1, 0)
                                .basal_dendrite(1, 2, 0).up()
                                .basal_dendrite(3, 2, 0).up(2)
                            .axon(2, -1, 0)
                                .axon(1.5, -2, 0).up()
                                .axon(2.5, -2, 0).up(4)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 3, 0).up()
                            .apical_dendrite(1, 3, 0)

                .build()
        )


        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(max_branch_order, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["axon.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["dendrite.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.max_branch_order"], 2)




from typing import TypeVar, Type, Dict
import inspect

import warnings

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.constants import (
    SOMA, AXON, BASAL_DENDRITE, APICAL_DENDRITE)


Mr = TypeVar("Mr", bound="Mark")

class Mark:
    """ A tag, intended for use in feature selection.
    """
    @classmethod
    def validate(cls, data: Data) -> bool:
        """ Determine if this feature is calculable from the provided data.

        Parameters
        ----------
        data : Data from a single morphological reconstruction

        Returns
        -------
        whether marked features can be calculated from these data

        """

        return True

    @classmethod
    def factory(cls: Type[Mr], name: str) -> Type[Mr]:
        return type(name, (cls,), {})


class RequiresLayerAnnotations(Mark):
    @classmethod
    def validate(cls, data: Data) -> bool:
        """ Checks whether each node in the data's morphology is annotated with
        a cortical layer. Returns False if any are missing.
        """
        return check_nodes_have_key(data, "layer")


class Intrinsic(Mark):
    """Indicates intrinsic features that don't rely on a ccf or scale."""
    pass


class Geometric(Mark):
    """Indicates features that change depending on coordinate frame."""
    pass


class AllNeuriteTypes(Mark):
    """Indicates features that are calculated for all neurite types."""
    pass


class RequiresDendrite(Mark):
    """This feature can only be calculated for neurons with at least one 
    dendrite node"""

    @classmethod
    def validate(cls, data: Data) -> bool:
        return data.morphology.has_type(APICAL_DENDRITE) \
            or data.morphology.has_type(BASAL_DENDRITE)

class RequiresRelativeSomaDepth(Mark):
    """This feature can only be calculated for relative soma depth"""

    @classmethod
    def validate(cls, data: Data) -> bool:
        return data.morphology.has_type(RELATIVE_SOMA_DEPTH)

class RequiresSoma(Mark):
    """Indicates that these features require a soma."""
    @classmethod
    def validate(cls, data: Data) -> bool:
        return data.morphology.has_type(SOMA)


class RequiresApical(Mark):
    """Indicates that these features require an apical dendrite."""
    @classmethod
    def validate(cls, data: Data) -> bool:
        return data.morphology.has_type(APICAL_DENDRITE)


class RequiresBasal(Mark):
    """Indicates that these features require a basal dendrite."""
    @classmethod
    def validate(cls, data: Data) -> bool:
        return data.morphology.has_type(BASAL_DENDRITE)


class RequiresAxon(Mark):
    """Indicates that these features require an axon."""
    @classmethod
    def validate(cls, data: Data) -> bool:
        return data.morphology.has_type(AXON)


# TODO: this describes the present requirements of root-dependent features. I 
# think that in nearly every case we actually want a RequiresUniqueSomaNode 
# mark
class RequiresRoot(Mark):
    """Indicates that this features require a root. Warns if the root 
    is not unique"""

    @classmethod
    def validate(cls, data: Data) -> bool:
        num_roots = len(data.morphology.get_roots())

        if num_roots > 1:
            warnings.warn(
                f"This morphology is not uniquely rooted! Found {num_roots} "
                "root nodes. Features using the root node of this morphology "
                "may not select that node consistently. Some or all of these "
                "root nodes may not be soma nodes."
            )
        elif num_roots < 1:
            return False

        return True


class BifurcationFeatures(Mark):
    """Indicates a feature calculated on bifurcations."""
    pass


class CompartmentFeatures(Mark):
    """Indicates a feature calculated on compartments."""
    pass


class TipFeatures(Mark):
    """Indicates a feature calculated on tips (leaf nodes)."""
    pass


class NeuriteTypeComparison(Mark):
    """Indicates a feature that is a comparison between neurite types.

    Function should be decorated with the appropriate RequiresType marks
    """
    pass



class RequiresRadii(Mark):
    """ This feature can only be calculated if the radii of nodes are annotated.
    """

    @classmethod
    def validate(cls, data: Data) -> bool:
        return check_nodes_have_key(data, "radius")


class RequiresReferenceLayerDepths(Mark):
    """ This feature can only be calculated if a referenceset of average depths 
    for cortical layers is provided. See features.layer.reference_layer_depths
    for more information.
    """

    @classmethod
    def validate(cls, data: Data) -> bool:
        return hasattr(data, "reference_layer_depths")


class RequiresLayeredPointDepths(Mark):
    """ This feature can only be calculated if (cortical) points are annotated 
    with a collection of within-layer depths. See 
    features.layer.layered_point_depths for more information.
    """

    @classmethod
    def validate(cls, data: Data) -> bool:
        return hasattr(data, "layered_point_depths")



class RequiresRegularPointSpacing(Mark):
    """ This features can only be (meaningfully) calculated if the points (e.g.
    node positions) on which it is based are resampled to have regular spacing.
    """



def check_nodes_have_key(data: Data, key: str) -> bool:
    """ Checks whether each node in a morphology is annotated with some key.
    """

    has_key = True
    for node in data.morphology.nodes():
        has_key = has_key and key in node
        if not has_key:
            break

    return has_key



import marshmallow as mm
from argschema.schemas import ArgSchema, DefaultSchema
from argschema.fields import (
    InputFile, List, Float, Int, Nested
)
from neuron_morphology.transforms.affine_transformer._schemas import (
    AffineDictSchema
)


class InputParameters(ArgSchema):
    swc_path = InputFile(
        description='path to swc file',
        required=True
    )
    marker_path = InputFile(
        description='path to marker file',
        required=True
    )
    soma_depth = Float(
        description='Recorded Depth of soma in slice',
        required=True,
    )
    cut_thickness = Float(
        description='Thickness of the slice when cut',
        required=False,
        default=350.,
    )


class OutputParameters(DefaultSchema):
    inputs = Nested(
        InputParameters,
        description="The parameters argued to this executable",
        required=True
    )
    scale_correction = Float(
        description=(
            "Z axis scale factor to correct for slice shrinkage"
        ),
        required=True
    )
    scale_transform = Nested(AffineDictSchema,
                                  required=False,
                                  description='Dictionary defining an affine transform')



import sys
import os
import logging
import copy as cp

from typing import List, Tuple, Optional

import numpy as np
from scipy.interpolate import griddata
import xarray as xr

from argschema import ArgSchemaParser

from neuron_morphology.transforms.pia_wm_streamlines._schemas import (
    PiaWmStreamlineSchema, OutputParameters)
from neuron_morphology.transforms.streamline import generate_laplace_field


def convert_path_str_to_list(path_str: str,
                             resolution: float = 1.0
                             ) -> List[Tuple[float, float]]:

    vals = path_str.split(',')

    list_coords = []
    for x, y in zip(vals[0::2], vals[1::2]):
        list_coords.append((float(x) * resolution, float(y) * resolution))

    return list_coords


def run_streamlines(pia_path_str: str,
         wm_path_str: str,
         resolution: float,
         soma_path_str: Optional[str] = None,
         mesh_res: int = 20,
         pia_fixed_value: float = 1.0,
         wm_fixed_value: float = 0.0,):

    pia_path = convert_path_str_to_list(pia_path_str, resolution)
    wm_path = convert_path_str_to_list(wm_path_str, resolution)

    soma_center = np.asarray([0, 0])
    if soma_path_str:
        soma_path = convert_path_str_to_list(soma_path_str,
                                             resolution)
        soma_center = np.mean(soma_path, axis=0)
        pia_path = [(x - soma_center[0], y - soma_center[1])
                    for (x, y) in pia_path]
        wm_path = [(x - soma_center[0], y - soma_center[1])
                   for (x, y) in wm_path]

    (u, grad_u, mesh, mesh_coords, mesh_values, mesh_gradients) = \
        generate_laplace_field(pia_path,
                               wm_path,
                               mesh_res=mesh_res,
                               top_value=pia_fixed_value,
                               bottom_value=wm_fixed_value,
                               eps_bounds=1e-8)

    x = [coord[0] for coord in mesh_coords]
    y = [coord[1] for coord in mesh_coords]
    xx = np.arange(min(x)-1, max(x)+1, 1)
    yy = np.arange(min(y)-1, max(y)+1, 1)

    grid_x, grid_y = np.meshgrid(xx, yy, indexing='ij')
    grid_u = griddata((x, y),
                      np.asarray(mesh_values),
                      (grid_x, grid_y),
                      method='cubic')
    grid_grad_u = griddata((x, y),
                           np.asarray(mesh_gradients),
                           (grid_x, grid_y),
                           method='cubic')

    depth_field = xr.DataArray(
        data=grid_u,
        dims=['x', 'y'],
        coords={'x': xx, 'y': yy})
    gradient_field = xr.DataArray(
        data=grid_grad_u,
        dims=['x', 'y', 'dim'],
        coords={'x': xx, 'y': yy, 'dim': ['dx', 'dy']})

    translation = -soma_center

    return depth_field, gradient_field, translation


def main():
    parser = ArgSchemaParser(
        schema_type=PiaWmStreamlineSchema,
        output_schema_type=OutputParameters
    )

    args = cp.deepcopy(parser.args)
    logging.getLogger().setLevel(args.pop("log_level"))
    output_dir = args.pop('output_dir')
    output_json = args.pop('output_json')

    depth_field, gradient_field, translation = run_streamlines(**args)

    # save results to file
    depth_field_file = os.path.join(output_dir,
                                    'depth_field.nc')
    gradient_field_file = os.path.join(output_dir,
                                       'gradient_field.nc')
    depth_field.to_netcdf(depth_field_file)
    gradient_field.to_netcdf(gradient_field_file)

    output = {
        'inputs': parser.args,
        'translation': translation,
        'depth_field_file': depth_field_file,
        'gradient_field_file': gradient_field_file,
    }

    parser.output(output)

if __name__ == "__main__":
    main()



import gmsh
import ufl
import numpy as np
import dolfinx.fem as fem
from dolfinx.io import gmshio
from typing import List, Tuple
from mpi4py import MPI
from petsc4py.PETSc import ScalarType
from shapely import geometry as geo
from neuron_morphology.transforms.geometry import (
    get_ccw_vertices, get_vertices_from_two_lines
)


def solve_laplace_2d(V: fem.FunctionSpace,
                     domain,
                     bcs: List[fem.bcs.DirichletBCMetaClass]):
    """
        Solves the laplace equation with boundary conditions bcs on V

        Parameters
        ----------
        V: Fenics FunctionSpace object created from a mesh
        bcs: List of Fenics DirichletBC Boundary Conditions
    """

    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    f = fem.Constant(domain, ScalarType(0)) # no source for the Laplace equation
    a = ufl.dot(ufl.grad(u), ufl.grad(v)) * ufl.dx
    L = f * v * ufl.dx

    problem = fem.petsc.LinearProblem(a, L, bcs=bcs, petsc_options={"ksp_type": "preonly", "pc_type": "lu"})
    uh = problem.solve()

    return uh


def compute_gradient(uh, W, bcs=[]):
    Wf = fem.Function(W)
    f = ufl.grad(uh)
    V = Wf.function_space
    dx = ufl.dx(V.mesh)

    u = ufl.TrialFunction(V)
    v = ufl.TestFunction(V)
    a = ufl.inner(u, v) * dx
    L = ufl.inner(f, v) * dx

    problem = fem.petsc.LinearProblem(a, L, bcs=bcs, petsc_options={"ksp_type": "preonly", "pc_type": "lu"})
    grad_uh = problem.solve()

    return grad_uh


def generate_laplace_field(top_line: List[Tuple],
                           bottom_line: List[Tuple],
                           mesh_res: float = 20,
                           top_value: float = 1.0,
                           bottom_value: float = 0.0,
                           eps_bounds: float = 1e-8):
    """
        Solve Laplace equation inside a polygon bounded
        by two lines (top_line and bottom_line) and the artificial
        straight side lines connecting the ends of the two lines.
        Apply Dirichlet BC on top_line and bottom_line and
        zero Neuman BC on the side lines.


        Demo of fenics: https://github.com/hplgit/fenics-tutorial/
                        blob/master/pub/python/vol1/ft01_poisson.py

        If top_value and bottom_value defaults are used, value_field
        will be the normalized distance to the top_line

        Parameters
        ----------
        top_line: line that will have top_value boundary condition
        bottom_line: line that will have bottom_value boundary condition
        mesh_res: resolution of the mesh
        top_value: value for top Dirichlet Boundary
        bottom_value: value for bottom Dirichlet Boundary

        Returns
        -------
        u: returns value at input point e.g. 0.5 = u((0.5, 0.5))
        grad_u: returns gradient at input point e.g. [0, 1] = u((0.5, 0.5))

        mesh_coords: coordinates of each vertex in the mesh
        mesh_values: values at each vertex in the mesh
        gradient_mesh: gradient at each vertex in the mesh

    """

    # Collect and sort vertices for mesh generation
    circular_vertices = get_vertices_from_two_lines(top_line, bottom_line)
    vertices = get_ccw_vertices(circular_vertices)

    # Create mesh with gmsh
    gmsh.initialize()

    # Set up the points
    point_tags = [gmsh.model.geo.addPoint(x, y, 0, meshSize=mesh_res)
        for (x, y) in vertices[:-1]]

    # connect the points into lines
    line_tags = [gmsh.model.geo.addLine(point_tags[i], point_tags[i + 1])
        for i in range(len(point_tags) - 1)]
    line_tags += [gmsh.model.geo.addLine(point_tags[-1], point_tags[0])]

    # add the curve loop
    curve_loop = gmsh.model.geo.addCurveLoop(line_tags)

    # add the surface
    surface = gmsh.model.geo.addPlaneSurface([curve_loop])

    # generate the mesh
    gmsh.model.geo.synchronize()
    gdim = 2 # value of 2 for 2D
    gmsh.model.addPhysicalGroup(gdim, [surface])
    gmsh.model.mesh.generate(gdim)

    # import mesh into dolfinx
    gmsh_model_rank = 0
    mesh_comm = MPI.COMM_WORLD
    domain, cell_markers, facet_markers = gmshio.model_to_mesh(gmsh.model, mesh_comm, gmsh_model_rank, gdim=gdim)

    # Create variational space
    V = fem.FunctionSpace(domain, ("CG", 1))
    W = fem.VectorFunctionSpace(domain, ("CG", 1))

    # Create boundary conditions
    top_ls = geo.LineString(top_line)
    def boundary_top(coords):
        points = [geo.Point((x, y)) for x, y in coords[:2, :].T]
        return np.array([top_ls.distance(point) < eps_bounds for point in points])

    bottom_ls = geo.LineString(bottom_line)
    def boundary_bottom(coords):
        points = [geo.Point((x, y)) for x, y in coords[:2, :].T]
        return np.array([bottom_ls.distance(point) < eps_bounds for point in points])

    boundary_dofs_top = fem.locate_dofs_geometrical(V, boundary_top)
    boundary_dofs_bottom = fem.locate_dofs_geometrical(V, boundary_bottom)

    bc_top = fem.dirichletbc(ScalarType(top_value), boundary_dofs_top, V)
    bc_bottom = fem.dirichletbc(ScalarType(bottom_value), boundary_dofs_bottom, V)
    bcs = [bc_top, bc_bottom]

    # Solve for value field
    u = solve_laplace_2d(V, domain, bcs)

    # Get derivative
    grad_u = compute_gradient(u, W)  # grad_u = Grad

    # Get values at mesh points
    mesh_coords = domain.geometry.x[:, :2]
    mesh_values = u.x.array
    mesh_gradients = np.reshape(grad_u.x.array, (-1, 2))

    return u, grad_u, domain, mesh_coords, mesh_values, mesh_gradients



from argschema.schemas import ArgSchema, DefaultSchema
from argschema.fields import (
    InputFile, OutputFile, String, Nested, Dict, List, Int, Field, Float, Boolean)
from marshmallow import ValidationError


class PathResolution(DefaultSchema):
    path = String(
        description="Fixed value boundary condition",
        required=True
    )

    resolution = Float(
        description="Resolution of pixels in microns",
        required=True
    )


class PrimaryBoundaries(DefaultSchema):
    Soma = Nested(
        PathResolution,
        description='string alternating x, y coordinates outlining the soma. '
                    'If provided, streamlines will be translated so that '
                    'the origin is at the soma',
        required=True
    )

    White_Matter = Nested(
        PathResolution,
        description="Fixed value white matter boundary condition",
        required=False
    )

    Pia = Nested(
        PathResolution,
        description=("Fixed value pia boundary condition"),
        required=False
    )

class S3LandingBucket(DefaultSchema):
    name = String(
        description="s3 landing bucket name or access point arn",
        required=True
    )

    region = String(
        description="s3 landing bucket's region",
        required=True
    )

    credentials_file = String(
        description="the INI config file path to the s3 landing bucket's credentials",
        required=False
    )

class InputParameters(ArgSchema):

    destination_bucket = Nested(
        S3LandingBucket,
        description="s3 landing bucket info (bucket_name/access_point_arn and region)",
        required=True
    )

    neuron_reconstruction_id = Int(
        description="neuron reconstruction id",
        required=True,
    )

    specimen_id = Int(
        description="specimen id",
        required=True
    )

    primary_boundaries = Nested(
        PrimaryBoundaries,
        description="primary boundaries",
        required=True
    )

    swc_file = String(
        description="path to input swc (csv) file",
        required=True
    )

    cell_depth = Float(
        description="cell depth",
        required=True,
        allow_none=True
    )

    cut_thickness = Float(
        description="cut thickness",
        required=True,
        allow_none=True
    )

    marker_file = String(
        description="path to input marker (csv) file",
        required=True
    )

    ccf_soma_xyz = List(
        Float,
        cli_as_single_argument=True,
        description="soma location (x,y,z) coordinates in CCF",
        required=True
    )

    slice_transform = List(
        Float,
        cli_as_single_argument=True,
        description='List defining the transform defining slice cut angle',
        required=True,
        allow_none=True,
    )

    slice_image_flip = Boolean(
        description=('indicates whether the image was flipped relative '
                     'to the slice (avg_group_label.name = \'Flip Slice Indicator\''),
        required=True
    )



from tests.objects import test_node, test_tree
from neuron_morphology.validation.result import InvalidMorphology
from neuron_morphology.validation import structure_validation as sv
from tests.validation.validation_test_case import ValidationTestCase
from neuron_morphology.constants import *
import unittest
from mock import patch


class TestStructureValidationFunctions(ValidationTestCase):
    """ Tests the functions in structure_validation.py """

    @patch("neuron_morphology.validation.swc_validators", [sv])
    def test_children_nodes_appear_before_parent_nodes_valid(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=APICAL_DENDRITE, parent_node_id=1)]

        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [sv])
    def test_children_nodes_appear_before_parent_nodes_not_valid(self):
        try:
            nodes = [test_node(id=1, type=AXON, parent_node_id=2),
                     test_node(id=2, type=SOMA, parent_node_id=-1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Child node needs to come before parent node",
                                  [[1]])


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestStructureValidationFunctions)
    unittest.TextTestRunner(verbosity=5).run(suite)



import unittest

import numpy as np

from neuron_morphology.morphology import Morphology
from neuron_morphology.morphology_builder import MorphologyBuilder

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor
from neuron_morphology.feature_extractor.feature_specialization import (
    NEURITE_SPECIALIZATIONS)
from neuron_morphology.feature_extractor.marked_feature import specialize

from neuron_morphology.features import intrinsic as ic


class TestIntrinsic(unittest.TestCase):

    def setUp(self):

        # Create an axon that branches once,
        # and a basal dendrite that has a bifurcation and a trifurcation
        self.morphology = (
            MorphologyBuilder()
                .root()
                    .axon()
                        .axon()
                            .axon()
                                .axon().up()
                                .axon().up(4)
                    .basal_dendrite()
                        .basal_dendrite()
                            .basal_dendrite().up()
                            .basal_dendrite()
                                .basal_dendrite().up()
                                .basal_dendrite().up()
                                .basal_dendrite()
                .build()
        )

        self.data = Data(self.morphology)

        self.num_nodes = specialize(
            ic.num_nodes,
            NEURITE_SPECIALIZATIONS)
        self.num_tips = specialize(
            ic.num_tips,
            NEURITE_SPECIALIZATIONS)
        self.num_branches = specialize(
            ic.num_branches,
            NEURITE_SPECIALIZATIONS)
        self.mean_fragmentation = specialize(
            ic.mean_fragmentation,
            NEURITE_SPECIALIZATIONS)
        self.max_branch_order = specialize(
            ic.max_branch_order,
            NEURITE_SPECIALIZATIONS)

    def extract(self, feature):
        extractor = FeatureExtractor([feature])
        return (
            extractor.extract(self.data).results
        )

    def test_all_neurites_num_nodes(self):
        self.assertEqual(
            self.extract(self.num_nodes)['all_neurites.num_nodes'],
            13)

    def test_all_neurites_num_tips(self):
        self.assertEqual(
            self.extract(self.num_tips)['all_neurites.num_tips'],
            6)

    def test_all_neurites_num_branches(self):
        self.assertEqual(
            self.extract(self.num_branches)['all_neurites.num_branches'],
            10)

    def test_all_neurites_max_branch_order(self):
        self.assertEqual(
            self.extract(self.max_branch_order)['all_neurites.max_branch_order'],
            3)

    def test_all_neurites_mean_fragmentation(self):
        self.assertAlmostEqual(
            self.extract(self.mean_fragmentation)['all_neurites.mean_fragmentation'],
            13/10)  # 13 compartments in 10 branches

    def test_number_of_branches_by_types(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(2, 1, 0)
                                .basal_dendrite(1, 2, 0).up()
                                .basal_dendrite(3, 2, 0).up(2)
                            .axon(2, -1, 0)
                                .axon(1.5, -2, 0).up()
                                .axon(2.5, -2, 0).up(4)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 3, 0).up()
                            .apical_dendrite(1, 3, 0)

                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(ic.num_branches, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.num_branches"], 10)
        self.assertEqual(feature_extraction_run.results["axon.num_branches"], 3)
        self.assertEqual(feature_extraction_run.results["dendrite.num_branches"], 6)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.num_branches"], 3)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.num_branches"], 3)

    def test_max_branch_order_by_types(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(2, 1, 0)
                                .basal_dendrite(1, 2, 0).up()
                                .basal_dendrite(3, 2, 0).up(2)
                            .axon(2, -1, 0)
                                .axon(1.5, -2, 0).up()
                                .axon(2.5, -2, 0).up(4)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 3, 0).up()
                            .apical_dendrite(1, 3, 0)

                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(ic.max_branch_order, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["axon.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["dendrite.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.max_branch_order"], 3)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.max_branch_order"], 2)




"""Contains a simple utility for representing a "growable" 2D rectangle
"""
from typing import Callable

import numpy as np

from neuron_morphology.snap_polygons.types import TransformType


class BoundingBox:
    """ Represents the bounds of a set of 2D objects

    Parameters
    ----------
    vorigin, horigin : the near corner of the box
    vextent, hextent : the far corner of the box

    """

    __slots__ = ["vorigin", "horigin", "vextent", "hextent"]

    # one might think these should be int, but we want to be able to record 
    # precisely the actual bounding box of some polygons with float vertices 
    # (then scale and convert to int at a requested resolution when necessary)
    def __init__(
            self, 
            horigin: float, 
            vorigin: float, 
            hextent: float,
            vextent: float, 
    ):

        self.vorigin = vorigin
        self.horigin = horigin

        self.vextent = vextent
        self.hextent = hextent

    def __repr__(self):
        return (
            f"BoundingBox({self.horigin}, {self.vorigin}, "
            f"{self.hextent}, {self.vextent})"
        )

    @property
    def width(self) -> float:
        """Horizontal side length of the box.
        """
        return self.hextent - self.horigin

    @property
    def height(self) -> float:
        """Vertical side length of the box.
        """
        return self.vextent - self.vorigin

    @property
    def aspect_ratio(self) -> float:
        """Width / height ratio of the box.
        """
        return self.width / self.height 
    
    @property
    def origin(self):
        """Coordinates of the box's origin.
        """
        return [self.horigin, self.vorigin]

    @property
    def extent(self):
        """Coordinates of the box's extent (opposite corner from the origin).
        """
        return [self.hextent, self.vextent]

    @property
    def coordinates(self):
        """Origin and extent coordinates.
        """
        return [self.horigin, self.vorigin, self.hextent, self.vextent]

    def update(
            self, 
            horigin: float, 
            vorigin: float, 
            hextent: float,
            vextent: float, 
    ):
        """ Potentially enlarges this box.

        Parameters
        ----------
        As to the constructor of BoundingBox. The new shape of this box is the 
            smallest box enclosing both this and the inputs.

        """

        self.vorigin = min(self.vorigin, vorigin)
        self.horigin = min(self.horigin, horigin)

        self.vextent = max(self.vextent, vextent)
        self.hextent = max(self.hextent, hextent)

    def transform(
            self, 
            transform: TransformType, 
            inplace: bool = False
    ) -> "BoundingBox":
        """ Apply a transform to this box

        Parameters
        ----------
        transform : A callable which maps (vertical, horizontal) coordinates to 
            new (vertical, horizontal) coordinates.
        inplace : if True, apply the transform to this object

        Returns
        -------
        The transformed box (potentially self)

        """

        if inplace:
            obj = self
        else:
            obj = self.copy()

        obj.horigin, obj.vorigin = transform(
            obj.horigin, obj.vorigin)
        obj.hextent, obj.vextent = transform(
            obj.hextent, obj.vextent)
        
        return obj

    def copy(self) -> "BoundingBox":
        """ Duplicates this bounding box

        Returns
        -------
        A copy of this object.

        """

        return self.__class__(
            self.horigin, self.vorigin, 
            self.hextent, self.vextent
        )

    def round(
            self, 
            inplace: bool = False, 
            origin_via: Callable[[float], float] = np.around,
            extent_via: Callable[[float], float] = np.around
    ):
        """ Round the coordinates of this box

        Parameters
        ----------
        inplace : If True, round the coordinates of this object
        origin_via : method to use when rounding the origin
        extent_via : method to use when rounding the extent

        Returns
        -------
        The rounded box (potentially self.

        """

        if inplace:
            obj = self
        else:
            obj = self.copy()

        obj.vorigin = origin_via(obj.vorigin)
        obj.horigin = origin_via(obj.horigin)

        obj.vextent = extent_via(obj.vextent)
        obj.hextent = extent_via(obj.hextent)

        return obj



from operator import add
from . import bits_validation as bv
from . import marker_validation as mv
from . import radius_validation as rv
from . import resample_validation as rev
from . import type_validation as tv
from . import structure_validation as stv
from functools import reduce


swc_validators = [bv, rv, rev, tv, stv]
marker_validators = [mv]


def validate_morphology(morphology):

    result = reduce(add, (sv.validate(morphology) for sv in swc_validators))

    return result


def validate_marker(marker, morphology):

    result = reduce(add, (m.validate(marker, morphology) for m in marker_validators))

    return result



import unittest

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.marked_feature import (
    specialize
)
from neuron_morphology.feature_extractor.feature_specialization import (
    BasalDendriteSpec
)
from neuron_morphology.features.soma import (
    calculate_number_of_stems
)

class TestNumberOfStems(unittest.TestCase):

    def test_number_of_stems_with_type(self):

        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .axon().up()
                    .basal_dendrite().up()
                    .basal_dendrite()
                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(calculate_number_of_stems, [BasalDendriteSpec])])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(
            feature_extraction_run.results["basal_dendrite.calculate_number_of_stems"],
            2
        )





from typing import (
    NamedTuple, Set, Dict, Any, Sequence, Type, List, TypeVar, Optional)
import abc
import copy as cp

from neuron_morphology.constants import (
    AXON,
    APICAL_DENDRITE,
    BASAL_DENDRITE,
    SOMA
)

from neuron_morphology.feature_extractor.mark import (
    Mark, 
    RequiresAxon,
    RequiresApical,
    RequiresBasal,
    RequiresDendrite,
    AllNeuriteTypes
)


Fs = TypeVar("Fs", bound="FeatureSpecialization")


class FeatureSpecialization(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def marks(self) -> Set[Type[Mark]]:
        pass

    @property
    @abc.abstractmethod
    def kwargs(self) -> Dict[str, Any]:
        pass

    @classmethod
    def factory(
        cls: Type[Fs], 
        name: str, 
        marks: Set[Type[Mark]], 
        kwargs: Dict[str, Any], 
        display_name: Optional[str] = None
    ) -> Type[Fs]:
        """ A utility for quickly generating feature specializations

        Parameters
        ----------
        name : The name of the generated class. If display_name is not 
            provided, this will also be used as the name attribute of the 
            generated class
        marks : the marks which this specialization implies.
        kwargs : the keyword argument values defining this specialization
        display_name : if provided, the name attribute of the generated 
            specialization.

        Returns
        -------
        A generated FeatureSpecialization subclass

        """

        if display_name is None:
            display_name = name

        return type(
            name,
            (cls,),
            {
                "name": display_name,
                "marks": cp.deepcopy(marks),
                "kwargs": cp.deepcopy(kwargs)
            }
        )


SpecializationOption = Type[FeatureSpecialization]
SpecializationSet = Set[SpecializationOption]
SpecializationSets = List[SpecializationSet]


class AxonSpec(FeatureSpecialization):
    name="axon"
    marks={RequiresAxon}
    kwargs={"node_types": [AXON]}

class ApicalDendriteSpec(FeatureSpecialization):
    name="apical_dendrite"
    marks={RequiresApical}
    kwargs={"node_types": [APICAL_DENDRITE]}

class BasalDendriteSpec(FeatureSpecialization):
    name="basal_dendrite"
    marks={RequiresBasal}
    kwargs={"node_types": [BASAL_DENDRITE]}

class DendriteSpec(FeatureSpecialization):
    name="dendrite"
    marks={RequiresDendrite}
    kwargs={"node_types": [APICAL_DENDRITE, BASAL_DENDRITE]}

class AllNeuriteSpec(FeatureSpecialization):
    name="all_neurites"
    marks={AllNeuriteTypes}
    kwargs={"node_types": None}


NEURITE_SPECIALIZATIONS = {
    AxonSpec,
    ApicalDendriteSpec,
    BasalDendriteSpec,
    DendriteSpec,
    AllNeuriteSpec
}

class AxonCompareSpec(FeatureSpecialization):
    name="axon"
    marks={RequiresAxon}
    kwargs={"node_types_to_compare": [AXON]}

class ApicalDendriteCompareSpec(FeatureSpecialization):
    name="apical_dendrite"
    marks={RequiresApical}
    kwargs={"node_types_to_compare": [APICAL_DENDRITE]}

class BasalDendriteCompareSpec(FeatureSpecialization):
    name="basal_dendrite"
    marks={RequiresBasal}
    kwargs={"node_types_to_compare": [BASAL_DENDRITE]}

class DendriteCompareSpec(FeatureSpecialization):
    name="dendrite"
    marks={RequiresDendrite}
    kwargs={"node_types_to_compare": [APICAL_DENDRITE, BASAL_DENDRITE]}

class AllNeuriteCompareSpec(FeatureSpecialization):
    name="all_neurites"
    marks={AllNeuriteTypes}
    kwargs={"node_types_to_compare": None}


NEURITE_COMPARISON_SPECIALIZATIONS = {
    AxonCompareSpec,
    ApicalDendriteCompareSpec,
    BasalDendriteCompareSpec,
    DendriteCompareSpec,
    AllNeuriteCompareSpec
}


import os
import sys
import subprocess as sp
import unittest
import shutil
import tempfile

import xarray as xr
import numpy as np

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.swc_io import morphology_to_swc
from neuron_morphology.transforms.upright_angle.compute_angle import (
    get_upright_angle)
from neuron_morphology.transforms.affine_transform import AffineTransform

import allensdk.core.json_utilities as ju

class TestUprightAngle(unittest.TestCase):
    def setUp(self):
        # input gradient in xarray format
        # up is negative x, so expected rotation is - pi / 2
        dx = -np.ones((6, 6))
        dy = np.zeros((6, 6))

        x = np.array([-35, -20, -5, 5, 20, 35])
        y = np.array([-35, -20, -5, 5, 20, 35])

        grad_data = np.empty((6,6,2))

        grad_data[:,:,0] = dx
        grad_data[:,:,1] = dy

        self.morphology = (
            MorphologyBuilder()
                .root(1, 2, 3)
                    .axon(0,2,3)
                        .build()
        )
        self.gradient = xr.DataArray(grad_data, dims=['x','y','dim'], coords={'x': x, 'y': y, 'dim': ['dx','dy']})

        self.test_dir = tempfile.mkdtemp()
        self.gradient_path = os.path.join(self.test_dir, 'gradient.nc')
        self.gradient.to_netcdf(self.gradient_path)

        self.swc_path = os.path.join(self.test_dir, '123.swc')
        morphology_to_swc(self.morphology, self.swc_path)

        self.output_json_path = os.path.join(self.test_dir, 'output.json')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_upright_angle(self):
        upright_angle = get_upright_angle(self.gradient, [0,0,0])
        self.assertAlmostEqual(upright_angle, -np.pi / 2, 3)

    def test_upright_end_to_end(self):
        input = {'gradient_path': self.gradient_path,
                 'node': '[0, 0, 0]',
                 'swc_path': self.swc_path,
                 'output_json': self.output_json_path}
        cmd = ['python', '-m',
               'neuron_morphology.transforms.upright_angle.compute_angle']
        for key, value in input.items():
            cmd.append(f'--{key}')
            cmd.append(f'{value}')

        sp.check_call(cmd)

        outputs = ju.read(self.output_json_path)
        self.assertAlmostEqual(outputs['upright_angle'], -np.pi / 2, 3)

        aff_t = AffineTransform.from_dict(outputs['upright_transform_dict'])
        morph_t = aff_t.transform_morphology(self.morphology)

        axon = morph_t.node_by_id(1)
        self.assertAlmostEqual(axon['x'], 1)
        self.assertAlmostEqual(axon['y'], 3)
        self.assertAlmostEqual(axon['z'], 3)



from typing import Dict, Optional, List, Any

import logging
import copy as cp

import numpy as np
from scipy.interpolate import interp2d
import xarray as xr

from argschema.argschema_parser import ArgSchemaParser
from neuron_morphology.transforms.upright_angle._schemas import (
    InputParameters, OutputParameters)
from neuron_morphology.morphology import Morphology
from neuron_morphology.swc_io import morphology_from_swc
import neuron_morphology.transforms.affine_transform as aff


def get_upright_angle(gradient: xr.DataArray,
                      point: Optional[List[float]] = None,
                      n_win: int = 2) -> float:

    """
        Calculate the upright angle at a position, e.g. soma, given a vector field

        Parameters
        ----------
        gradient: xarray of the the vector field
        point: list [x,y,z] coordinates
        n_win: number of grid points to define the interpolation window

        Returns
        -------
        angle

    """

    if not point:
        point = [0, 0, 0]

    x_point, y_point, z_point = point

    nx_idx = np.searchsorted(gradient.x, x_point)
    ny_idx = np.searchsorted(gradient.y, y_point)

    # Only use the  n_win points on each side, because if
    # the full array has any nans, interp2d will return nan

    x_win = gradient.x[nx_idx - n_win:nx_idx + n_win]
    y_win = gradient.y[ny_idx - n_win:ny_idx + n_win]

    values_win = gradient[
             nx_idx - n_win:nx_idx + n_win,
             ny_idx - n_win:ny_idx + n_win,
             :]

    # Also transpose because if x=m,y=n, interp2d requires z=(n,m)
    f_dx = interp2d(x_win, y_win, values_win[:, :, 0].T)
    f_dy = interp2d(x_win, y_win, values_win[:, :, 1].T)

    dx = f_dx(x_point, y_point)
    dy = f_dy(x_point, y_point)

    return np.pi / 2 - np.arctan2(dy[0], dx[0])

def calculate_transform(gradient_field: xr.DataArray,
         morph: Morphology,
         node: Optional[List[float]] = None,
                      ):
    theta = get_upright_angle(gradient_field, node)
    transform = np.eye(4)
    transform[0:3, 0:3] = aff.rotation_from_angle(theta)

    soma = morph.get_soma()

    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)

    transform[0:3,3] = np.asarray([
        -soma["x"] * cos_theta + soma["y"] * sin_theta + soma["x"],
        -soma["x"] * sin_theta - soma["y"] * cos_theta + soma["y"],
        0
    ])

    output = {
        'upright_transform': aff.AffineTransform(transform),
        'upright_angle': theta
    }

    return output

def run_upright_angle(gradient_path: str,
         swc_path: str,
         node: Optional[List[float]] = None,
                      ):

    try:
        gradient_field = xr.open_dataarray(gradient_path)
    except IOError:
        raise IOError(f"Cannot find file with the gradient field in {gradient_path}")

    morph = morphology_from_swc(swc_path)

    output = calculate_transform(gradient_field, morph, node)
    output["upright_transform_dict"] = output.pop("upright_transform").to_dict()

    return output


def main():
    parser = ArgSchemaParser(
        schema_type=InputParameters,
        output_schema_type=OutputParameters
    )

    args = cp.deepcopy(parser.args)
    logging.getLogger().setLevel(args.pop("log_level"))

    output = run_upright_angle(
        args["gradient_path"],
        args["swc_path"],
        args["node"],
    )
    output.update({"inputs": parser.args})

    parser.output(output)


if __name__ == "__main__":
    main()



import pytest
import os
import boto3
import zipfile
from moto import mock_s3
import unittest

from neuron_morphology.pipeline.post_data_to_s3 import post_object_to_s3, zip_files

path_to_current_file = os.path.realpath(__file__)
current_directory = os.path.split(path_to_current_file)[0]

file_name1 = "test_post_data_to_s3.py"
file_name2 = "__init__.py"

file_path1 = os.path.join(current_directory, file_name1)
file_path2 = os.path.join(current_directory, file_name2)

file_dict = {file_name1: file_path1, file_name2: file_path2}
bucket_name = "neuronmorphologypipeline"
archive_name = "test_post_data.zip"
region = "us-west-2"

class TestZipFiles(unittest.TestCase):
    def setUp(self):
        self.mock_listing = [file_name1, file_name2]
        self.file_dict = file_dict

    def test_zip_files(self):
        
        archive = zip_files(self.file_dict)

        if zipfile.is_zipfile(archive): 
            file_list = [file_name for file_name in zipfile.ZipFile(archive).namelist()]

        self.assertListEqual(self.mock_listing, file_list)

@mock_s3
def test_post_data_to_s3(tmpdir_factory):
    
    temp_output_dir = str(tmpdir_factory.mktemp("test"))
    temp_output_file = os.path.join(temp_output_dir, archive_name)

    s3 = boto3.resource('s3', region_name=region)
    s3.create_bucket(Bucket=bucket_name,  CreateBucketConfiguration={"LocationConstraint": region})

    archive_data = zip_files(file_dict)
    post_object_to_s3(archive_data.getvalue(), archive_name, bucket_name, region)
        
    s3_client = boto3.client('s3', region)

    s3_client.download_file(bucket_name, archive_name, temp_output_file)

    archive = zipfile.ZipFile(temp_output_file, 'r')

    assert all([a.filename == b for a, b in zip(archive.infolist(), [file_name1, file_name2])])



""" This module contains utilities for processing cortical surface drawings. 
In general we take these as given (they even take precedence of e.g. the 
upper and lower surfaces of layers 1 and 6b for instance), but some drawings 
pose resolvable problems.

The main such problem occurs when cortical layer drawings extend far from the 
layer drawings. Extrapolating layer drawings into this space is dangerous and 
not very useful (only the drawings near the cell are useful downstream). The 
solution implemented here is to cut out a segment of each surface whose 
endpoints are sufficiently close to the layer drawings and discard the rest.
"""
from typing import Union, Tuple, Callable, Sequence
import copy as cp
import logging

from shapely.geometry import LineString, Point
from shapely.geometry.base import BaseGeometry

from neuron_morphology.snap_polygons.types import LineType, ensure_linestring


ConditionFn = Callable[[Point], bool]


def trim_to_close(
        geometry: BaseGeometry, 
        threshold: float, 
        linestring: LineType,
        iterations: int = 10
) -> LineString:
    """Find the longest segment of a linestring whose endpoints are within a 
    specified distance of a geometry.

    Parameters
    ----------
    geometry : Acceptable distances are defined as extending from this object.
    threshold : Acceptable distances are less than or equal to this value
    linestring : to be trimmed (not in place)
    iterations : Use this many iterations to refine the endpoints of the 
        linestring

    Returns
    -------
    a trimmed copy of the input linestring

    """

    linestring = ensure_linestring(cp.deepcopy(linestring))

    def condition(point: Point) -> bool:
        return geometry.distance(Point(point)) <= threshold

    try:
        coords = trim_coords(
            linestring.coords, 
            condition=condition,
            iterations=iterations,
        )
    except ValueError:
        logging.error("no point within %s of argued geometry", threshold)
        raise

    return LineString(coords)


def find_transition(
        unmet: Point, 
        met: Point, 
        condition: ConditionFn, 
        iterations: int
) -> Point:
    """Given two points in space, one of which meets a condition, locate the 
    position along a line segment between these points where the condition 
    becomes true.

    Parameters
    ----------
    unmet : a point at which the condition is not met
    met : a point at which the condition is met
    condition : used to evaluate intermediate points
    iterations : refine this many times 

    Returns
    -------
    A point along the input segment at which the condition is met.

    Notes
    -----
    No such transition point is required to exist. In that case, this function 
    will find an arbitrary condition-meeting point along the segment. For our 
    use case, this misbehavior is tolerable because an exact transition point 
    is not required.
    """

    segment = LineString([unmet, met])
    midpoint = segment.interpolate(segment.length / 2)

    if condition(midpoint):
        met = midpoint
    else:
        unmet = midpoint

    if iterations > 0:
        return find_transition(unmet, met, condition, iterations - 1)
    return met
   

def first_met(
        coords: Sequence[Union[Point, Tuple]], 
        condition: ConditionFn, 
        iterations: int
) -> Tuple[int, Point]:
    """Locate the first point along a coordinate sequence at which a condition 
    is met.

    Parameters
    ----------
    coords : sequence to evaluate
    condition : used to evaluate points
    iterations : how many times to refine the transition point.

    Returns
    -------
    The index and value of the transition point.
    """
    
    for idx, coord in enumerate(coords):
        coord = Point(coord)
        met = condition(coord)

        if met:
            if idx == 0:
                return idx, coord
            return idx, find_transition(
                coords[idx - 1], coord, condition, iterations
            )

    raise ValueError("condition never met!")


def remove_duplicates(coords: Sequence[Point]) -> Sequence[Point]:
    """Remove duplicate points from a coordinate sequence.

    Parameters
    ----------
    coords : sequence with potential duplicates

    Returns
    -------
    list of coordinates with duplicates removed
    """
    if len(coords) < 2:
        return coords

    out = [coords[0]]
    for coord in coords[1:]:
        coord = Point(coord)
        if coord != out[-1]:
            out.append(coord)

    return out


def trim_coords(
        coords: Sequence[Union[Point, Tuple]], 
        condition: ConditionFn, 
        iterations: int
):  
    """Find the longest subinterval of a coordinate sequence whose endpoints 
    meet some condition.

    Parameters
    ----------
    coords : sequence to trim
    condition : used to evaluate points
    iterations : how many times to refine the endpoints.

    Returns
    -------
    Trimmed sequence
    """

    left_index, left_pt = first_met(coords, condition, iterations)
    right_index, right_pt = first_met(coords[::-1], condition, iterations)

    if right_index > 0:
        coords = list(coords[:-right_index]) + [right_pt]
    if left_index > 0:
        coords = [left_pt] + list(coords[left_index:])

    return remove_duplicates(coords)



import os
import subprocess as sp
import unittest
import shutil
import tempfile

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.swc_io import morphology_to_swc
from neuron_morphology.transforms.scale_correction.compute_scale_correction import (
    estimate_scale_correction, run_scale_correction)
from neuron_morphology.transforms.affine_transform import AffineTransform

import allensdk.core.json_utilities as ju


class TestScaleCorrection(unittest.TestCase):
    def setUp(self):

        self.morphology = (
            MorphologyBuilder()
                .root(1, 2, 3, radius=1.3)
                    .axon(0, 2, 100, radius=0.4)
                        .build()
        )
        self.max_morphology = (
            MorphologyBuilder()
                .root(1, 2, 3)
                    .axon(0, 2, 353)
                        .build()
        )
        self.soma_depth = 1.0
        self.cut_thickness = 350

        self.test_dir = tempfile.mkdtemp()

        self.swc_path = os.path.join(self.test_dir, '123.swc')
        morphology_to_swc(self.morphology, self.swc_path)

        self.marker_path = os.path.join(self.test_dir, '123.marker')
        with open(self.marker_path, 'w') as f:
            f.write("##x,y,z,radius,shape,name,comment, color_r,color_g,color_b\n"
                    "1.0,2.0,3.5,0.0,0,30,0,255, 255, 255\n"
                    "0.0,2.0,3.0,0.0,0,20,0,255, 255, 255")

        self.soma_marker_z = 3.5
        self.output_json_path = os.path.join(self.test_dir, 'output.json')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_scale_correction(self):
        scale = estimate_scale_correction(self.morphology,
                                     self.soma_depth,
                                     self.soma_marker_z,
                                     self.cut_thickness)
        self.assertAlmostEqual(scale, 2.0)
    def test_scale_correction_max(self):
        scale = estimate_scale_correction(self.max_morphology,
                                     self.soma_depth,
                                     self.soma_marker_z,
                                     self.cut_thickness)
        self.assertAlmostEqual(scale, 1.0)

    def test_scale_correction_end_to_end(self):
        input = {'swc_path': self.swc_path,
                 'marker_path': self.marker_path,
                 'soma_depth': self.soma_depth,
                 'cut_thickness': self.cut_thickness,
                 'output_json': self.output_json_path}
        cmd = ['python', '-m',
               'neuron_morphology.transforms.scale_correction.compute_scale_correction']
        for key, value in input.items():
            cmd.append(f'--{key}')
            cmd.append(f'{value}')

        sp.check_call(cmd)

        outputs = ju.read(self.output_json_path)
        self.assertAlmostEqual(outputs['scale_correction'], 2.0)

        aff_t = AffineTransform.from_dict(outputs['scale_transform'])
        morph_t = aff_t.transform_morphology(self.morphology,
                                             scale_radius=False)

        axon = morph_t.node_by_id(1)
        self.assertAlmostEqual(axon['x'], 0)
        self.assertAlmostEqual(axon['y'], 2)
        self.assertAlmostEqual(axon['z'], 200)
        self.assertAlmostEqual(axon['radius'], 0.4)

    def test_run_scale_correction(self):

        inputs = {
            "morphology": self.morphology,
            "soma_marker_z": self.soma_marker_z,
            "soma_depth": self.soma_depth,
            "cut_thickness": self.cut_thickness,
        }

        expected = {
            'scale_transform':
                {'tvr_00': 1.0,
                 'tvr_01': 0.0,
                 'tvr_02': 0.0,
                 'tvr_03': 0.0,
                 'tvr_04': 1.0,
                 'tvr_05': 0.0,
                 'tvr_06': 0.0,
                 'tvr_07': 0.0,
                 'tvr_08': 2.0,
                 'tvr_09': 0.0,
                 'tvr_10': 0.0,
                 'tvr_11': 0.0,
                 },
            'scale_correction': 2.0
        }

        obtained = run_scale_correction(**inputs)
        obtained.pop("morphology_scaled")

        print(expected)
        print(obtained)
        assert expected == obtained



import unittest
from tests.objects import test_morphology_large
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor
from neuron_morphology.features.default_features import default_features
from neuron_morphology.feature_extractor.data import Data


class TestDefault(unittest.TestCase):

    fe = FeatureExtractor()

    for feature in default_features:
        fe.register_features([feature])

    data = Data(test_morphology_large())
    feature_extraction_run = fe.extract(data)
    results = feature_extraction_run.results

    pass



import copy as cp
import logging
import multiprocessing as mp
import functools
from typing import Dict, Any, Tuple, List, Set, Optional, Type

from argschema import ArgSchemaParser

from neuron_morphology.feature_extractor._schemas import (
    InputParameters, OutputParameters)

from neuron_morphology.feature_extractor.run_feature_extraction import \
    run_feature_extraction

from neuron_morphology.feature_extractor.feature_writer import (
    FeatureWriter, DEFAULT_FEATURE_FORMATTERS)


def extract_multiple(
    reconstructions: List[Dict[str, Any]], 
    feature_set: str,
    heavy_output_path: str,
    required_marks: Optional[List[str]] = None,
    only_marks: Optional[List[str]] = None,
    num_processes: Optional[int] = None,
    global_parameters: Optional[Dict[str, Any]] = None,
    output_table_path: Optional[str] = None
):
    """ For each path in swc_paths, load the file into a morphology and (attempt 
    to) extract each feature in the set specified by feature_set.

    Because of how Windows handles multiprocessing, run_feature_extraction
    must be in another py file.

    Parameters
    ----------
    reconstructions : specify the reconstructions on which to compute features
    feature_set : names the set of features for which calculation will be
        attempted
    heavy_output_path : write "heavy" outputs, such as arrays, to this h5 file
    only_marks : names marks to which calculation will be restricted
    required_marks : raise an exception if these named marks fail validation
    num_processes : use this many cores in the multiprocessing pool.
    global_parameters : a dictionary specifying cross-reconstruction
        parameters
    output_table_path : if not none, write a flattened table of features here

    Returns
    -------
    a dictionary whose keys are reconstruction identifers and whose values are
        the outputs of run_feature_extraction for those reconstructions.

    """

    num_processes = num_processes if num_processes else mp.cpu_count()
    num_processes = min(num_processes, len(reconstructions))

    global_parameters = {} if global_parameters is None else global_parameters

    extract = functools.partial(
        run_feature_extraction,
        feature_set=feature_set,
        only_marks=only_marks,
        required_marks=required_marks,
        global_parameter_spec=global_parameters
    )

    if num_processes > 1:
        pool = mp.Pool(num_processes)
        mapper = pool.imap_unordered(extract, reconstructions)
    else:
        mapper = (extract(morph) for morph in reconstructions) # type: ignore[assignment]

    writer = FeatureWriter(
        heavy_output_path,
        output_table_path,
        formatters=DEFAULT_FEATURE_FORMATTERS
    )

    for identifier, run in mapper:
        writer.add_run(identifier, run)

    return writer.write()


def main():
    parser = ArgSchemaParser(
        schema_type=InputParameters,
        output_schema_type=OutputParameters
    )

    inputs_record = cp.deepcopy(parser.args)
    logging.getLogger().setLevel(inputs_record.pop("log_level"))
    inputs_record.pop("input_json", None)
    inputs_record.pop("output_json", None)
    output_table_path = inputs_record.pop("output_table_path", None)
    
    output = {}
    output.update({"inputs": parser.args})
    output.update({"results": extract_multiple(**inputs_record)})

    parser.output(output)


if __name__ == "__main__":
    main()


from typing import Optional, List

import numpy as np
from scipy import stats
from scipy.spatial import distance

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.features.statistics.coordinates import COORD_TYPE

from neuron_morphology.feature_extractor.marked_feature import marked
from neuron_morphology.feature_extractor.mark import Geometric,RequiresRoot


@marked(Geometric)
@marked(RequiresRoot)
def moments_along_max_distance_projection(
            data: Data,
            node_types: Optional[List] = None,
            coord_type: COORD_TYPE = COORD_TYPE.BIFURCATION,
            ):
    """
        Calculate the distance projections of a specific compartment and coordinate type along
        the line segment connecting soma to the most distant (from soma) node of that compartment.

        Parameters
        ----------
        data: Data Object containing a morphology
        node_types: a list of node types (see neuron_morphology constants)
        coord_type: Restrict which coordinate types are measured (i.e. projected along line segment)
                            (see neuron_morphology.features.statistics.coordinates for options)

        Returns
        -------
        summary_dict: summary stats of distances projected along specified line segment
    """

    morphology = data.morphology
    find_most_distant_coordinates = COORD_TYPE.NODE.get_coordinates(morphology, node_types=node_types)
    measuring_coordinates = coord_type.get_coordinates(morphology, node_types=node_types)

    if (not find_most_distant_coordinates) or (not measuring_coordinates):
        summary_dict = {
            'mean': np.nan,
            'std': np.nan,
            'var': np.nan,
            'skew': np.nan,
            'kurt': np.nan}

    else:
        measuring_coordinates = np.asarray(measuring_coordinates)
        soma_node = morphology.get_soma()
        soma_coord = np.array([soma_node['x'],soma_node['y'],soma_node['z']])

        max_distance = -1
        for coord in find_most_distant_coordinates:
            dist = distance.euclidean(coord,soma_coord)
            if dist>max_distance:
                max_distance = dist
                furthest_coord = coord

        projected_dists = np.dot( (furthest_coord - soma_coord).T, (measuring_coordinates - soma_coord).T )
        norm = np.linalg.norm(furthest_coord - soma_coord)

        distances = projected_dists/norm
        distances /= max_distance
        (_, _, mean, variance, skew, kurt) = stats.describe(distances, axis=0)
        stdv = np.std(distances)

        summary_dict = {
                'mean': mean,
                'std': stdv,
                'var': variance,
                'skew': skew,
                'kurt': kurt}

    return summary_dict




from tests.objects import test_node, test_tree
from neuron_morphology.validation import resample_validation as rev
from neuron_morphology.validation.result import InvalidMorphology
from tests.validation.validation_test_case import ValidationTestCase
from neuron_morphology.constants import *
import unittest
from mock import patch


class TestResampleValidationFunctions(ValidationTestCase):

    """ Tests the functions in resample_validation.py """

    @patch("neuron_morphology.validation.swc_validators", [rev])
    def test_distance_between_connected_nodes_valid(self):

        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=AXON, x=3188.34, y=2891.57, z=88.9906, parent_node_id=1),
                 test_node(id=3, type=AXON, x=3198.34, y=2888.57, z=89.9906, parent_node_id=2)]
        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [rev])
    def test_distance_between_connected_nodes_invalid(self):
        try:

            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=AXON, x=6725.2098, y=5890.6503, z=76.0, parent_node_id=1),
                     test_node(id=3, type=AXON, x=0, y=0, z=0, parent_node_id=2)]
            test_tree(nodes, strict_validation=True)

            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "The distance between two nodes should be less than 50px"
                                  , [[2, 3]])


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestResampleValidationFunctions)
    unittest.TextTestRunner(verbosity=5).run(suite)



import os
import subprocess as sp
import unittest
import shutil
import tempfile

import numpy as np
import h5py

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.swc_io import morphology_to_swc
from neuron_morphology.transforms.tilt_correction.compute_tilt_correction import (
    get_tilt_correction, run_tilt_correction, CCF_SHAPE, CCF_RESOLUTION)
from neuron_morphology.transforms.affine_transform import AffineTransform

import allensdk.core.json_utilities as ju


class TestTiltCorrection(unittest.TestCase):
    def setUp(self):

        self.morphology = (
            MorphologyBuilder()
                .root(1, 2, 3)
                    .axon(0, 2, 3)
                        .build()
        )

        self.test_dir = tempfile.mkdtemp()

        self.swc_path = os.path.join(self.test_dir, '123.swc')
        morphology_to_swc(self.morphology, self.swc_path)

        self.marker_path = os.path.join(self.test_dir, '123.marker')
        with open(self.marker_path, 'w') as f:
            f.write("##x,y,z,radius,shape,name,comment, color_r,color_g,color_b\n"
                    "1.0,2.0,2.5,0.0,0,30,0,255, 255, 255\n"
                    "0.0,2.0,3.0,0.0,0,20,0,255, 255, 255")

        self.slice_transform_list = [1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0]
        self.slice_transform = AffineTransform.from_list(self.slice_transform_list)
        self.slice_image_flip = True

        self.soma_marker = {'x': 1, 'y': 2, 'z': 0}
        self.soma_voxel = (400, 400, 1120)

        self.ccf_soma_location = [self.soma_voxel[0] * CCF_RESOLUTION,
                                  self.soma_voxel[1] * CCF_RESOLUTION,
                                  self.soma_voxel[2] * CCF_RESOLUTION
                                  ]
        self.csl_dict = dict(zip(['x', 'y', 'z'], self.ccf_soma_location))

        # Streamline is on z+, so should result in tilt of np.pi / 2
        self.closest_path = np.asarray([400 * np.ones((20,), dtype='i'),
                                        400 * np.ones((20,), dtype='i'),
                                        np.arange(1120, 1140, dtype='i')])

        self.ccf_path = os.path.join(self.test_dir, 'paths.h5')
        path = np.ravel_multi_index(self.closest_path, CCF_SHAPE)
        with h5py.File(self.ccf_path, 'w') as f:
            f.create_dataset("view lookup", (1, ), dtype='i', data=0)
            f.create_dataset("paths", (1, 20), dtype='i', data=path)

        self.output_json_path = os.path.join(self.test_dir, 'output.json')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_tilt_correction(self):
        tilt = get_tilt_correction(self.morphology,
                                   self.soma_voxel,
                                   self.slice_transform.affine,
                                   self.closest_path)
        self.assertAlmostEqual(tilt, -np.pi / 2)

    def test_run_tilt_correction(self):
        (tilt, transform) = run_tilt_correction(self.morphology,
                                                self.soma_marker,
                                                self.csl_dict,
                                                self.slice_transform,
                                                self.slice_image_flip,
                                                self.ccf_path)
        self.assertAlmostEqual(tilt, -np.pi / 2)

    def test_tilt_correction_end_to_end(self):
        input = {'swc_path': self.swc_path,
                 'marker_path': self.marker_path,
                 'ccf_soma_location': self.ccf_soma_location,
                 'slice_transform_list': self.slice_transform_list,
                 'slice_image_flip': self.slice_image_flip,
                 'ccf_path': self.ccf_path,
                 'output_json': self.output_json_path}
        cmd = ['python', '-m',
               'neuron_morphology.transforms.tilt_correction.compute_tilt_correction']
        for key, value in input.items():
            cmd.append(f'--{key}')
            cmd.append(f'{value}')

        sp.check_call(cmd)

        outputs = ju.read(self.output_json_path)
        self.assertAlmostEqual(outputs['tilt_correction'], -np.pi / 2)

        aff_t = AffineTransform.from_dict(outputs['tilt_transform_dict'])
        morph_t = aff_t.transform_morphology(self.morphology)

        axon = morph_t.node_by_id(1)
        self.assertAlmostEqual(axon['x'], 0)
        self.assertAlmostEqual(axon['y'], 3)
        self.assertAlmostEqual(axon['z'], -2)



from typing import List, Dict, Optional, Any

import numpy as np

from neuron_morphology.morphology import Morphology
from neuron_morphology.transforms.transform_base import TransformBase


class AffineTransform(TransformBase):
    """Handles transformations to a pia/wm aligned coordinate frame."""

    def __init__(self, affine: Optional[Any] = None):
        """
            Represent the transform as a (4,4) np.ndarray affine matrix

            Parameters
            ----------
            affine: (4,4) array-like affine transformation

        """
        if affine is None:
            self.affine = np.eye(4)
        else:
            self.affine = np.asarray(affine)
            assert self.affine.shape == (4, 4)

    @classmethod
    def from_dict(cls, affine_dict: Dict[str, float]):
        """
            Create an AffineTransform from a dict with keys and values.

            Parameters
            ----------
            affine_dict: keys and values corresponding to the following
                [[tvr_00 tvr_01 tvr_02 tvr_09]
                 [tvr_03 tvr_04 tvr_05 tvr_10]
                 [tvr_06 tvr_07 tvr_08 tvr_11]
                 [0      0      0      1]]

            Returns
            -------
            AffineTransform object

        """
        transform = np.reshape([affine_dict['tvr_%02d' % i] for i in range(9)],
                               (3, 3))
        translation = np.reshape([affine_dict['tvr_%02d' % i]
                                 for i in range(9, 12)], (3, 1))
        affine = affine_from_transform_translation(transform, translation)

        return cls(affine)

    @classmethod
    def from_list(cls, affine_list: List[float]):
        """
            Create an Affine Transform from a list

            Parameters
            ----------
            affine_list: list of tvr values corresponding to:
                [[tvr_00 tvr_01 tvr_02 tvr_09]
                 [tvr_03 tvr_04 tvr_05 tvr_10]
                 [tvr_06 tvr_07 tvr_08 tvr_11]
                 [0      0      0      1]]

            Returns
            -------
            AffineTransform object

        """
        transform = np.reshape(affine_list[0:9], (3, 3))
        translation = np.reshape(affine_list[9:12], (3, 1))
        affine = affine_from_transform_translation(transform, translation)

        return cls(affine)

    def to_dict(self) -> Dict:
        """
            Create dictionary defining the transformation.

            Returns
            -------
            Dict with keys and values corresponding to the following:

                [[tvr_00 tvr_01 tvr_02 tvr_09]
                 [tvr_03 tvr_04 tvr_05 tvr_10]
                 [tvr_06 tvr_07 tvr_08 tvr_11]
                 [0      0      0      1]]

        """
        affine_dict = {}
        for i in range(9):
            affine_dict['tvr_%02d' % i] = self.affine[0:3, 0:3].flatten()[i]

        for i in range(3):
            affine_dict['tvr_%02d' % (i + 9)] = \
                self.affine[0:3, 3].flatten()[i]

        return affine_dict

    def to_list(self) -> List:
        """
            Create a list defining the transformation.

            Returns
            -------
            List with values corresponding to the following:

                [[tvr_00 tvr_01 tvr_02 tvr_09]
                 [tvr_03 tvr_04 tvr_05 tvr_10]
                 [tvr_06 tvr_07 tvr_08 tvr_11]
                 [0      0      0      1]]

        """
        affine_list = self.affine[0:3, 0:3].flatten().tolist()
        affine_list += self.affine[0:3, 3].flatten().tolist()

        return affine_list

    def transform(self, vector: Any) -> np.ndarray:
        """
            Apply this transform to (3,) point or (n,3) array-like of points.

            Parameters
            ----------
            vector: a (3,) array-like point or a (n,3) array-like array
                    of points to be transformed

            Returns
            -------
            numpy.ndarray with same shape as input

        """
        vector = np.asarray(vector)
        dims = len(vector.shape)
        if dims == 1:
            vector = vector.reshape((3, 1))
        else:
            vector = vector.T

        vector = np.vstack((vector, np.ones((1, vector.shape[1]))))

        vec_t = np.dot(self.affine, vector)[0:3, :]
        if dims == 1:
            vec_t = vec_t.reshape((3,))
        else:
            vec_t = vec_t.T

        return vec_t

    def _get_scaling_factor(self) -> float:
        """
            Calculate the scaling factor from the affine matrix.

            Returns
            -------
            Scaling factor: 3rd root of the determinant.

        """
        determinant = np.linalg.det(self.affine)
        return np.power(abs(determinant), 1.0 / 3.0)

    def transform_morphology(self,
                             morphology: Morphology,
                             clone: bool = False,
                             scale_radius: bool = True,
                             ) -> Morphology:
        """
            Apply this transform to all nodes in a morphology.

            Parameters
            ----------
            morphology: a Morphology loaded from an swc file
            clone: make a new object if True
            scale_radius: apply radius scaling if True

            Returns
            -------
            A Morphology
        """
        if clone:
            morphology = morphology.clone()

        if scale_radius:
            scaling_factor = self._get_scaling_factor()
        else:
            scaling_factor = 1

        for node in morphology.nodes():
            coordinates = np.array((node['x'], node['y'], node['z']),
                                   dtype=float)
            new_coordinates = self.transform(coordinates)
            node['x'] = new_coordinates[0]
            node['y'] = new_coordinates[1]
            node['z'] = new_coordinates[2]
            # approximate with uniform scaling in each dimension
            node['radius'] *= scaling_factor

        return morphology


def affine_from_transform_translation(transform: Optional[Any] = None,
                                      translation: Optional[Any] = None,
                                      translate_first: bool = False):
    """
        Create affine from linear transformation and translation.

        Affine transformation of vector x -> Ax + b in 3D:
        [A,       b
         0, 0, 0, 1]
        A is a 3x3 linear tranformation
        b is a 3x1 translation

        Parameters
        ----------
        transform: linear transformation (3, 3) array-like
        translation: linear translation (3,) array-like
        translate_first: apply the translation before the transform

        Returns
        -------
        (4, 4) numpy.ndarray affine matrix

    """
    if transform is None:
        transform = np.eye(3)
    else:
        transform = np.asarray(transform)

    if translation is None:
        translation = np.zeros((3, 1))
    else:
        translation = np.reshape(translation, (3, 1))
        if translate_first:
            translation = transform.dot(translation)

    affine = np.zeros((4, 4))
    affine[0:3, 0:3] = transform
    affine[0:3, [3]] = translation
    affine[3, 3] = 1.0

    return affine


def rotation_from_angle(angle: float, axis: int = 2):
    """
        Create an affine matrix from a rotation about a specific axis.

        Parameters
        ----------
        angle: rotation angle in radians
        axis: axis to rotate about, 0=x, 1=y, 2=z (default z axis)

        Returns
        -------
        (3, 3) numpy.ndarray rotation matrix

    """
    c = np.cos(angle)
    s = np.sin(angle)

    if axis == 0:
        rot = np.asarray([[1, 0, 0],
                          [0, c, -s],
                          [0, s, c]])
    elif axis == 1:
        rot = np.asarray([[c, 0, s],
                          [0, 1, 0],
                          [-s, 0, c]])
    elif axis == 2:
        rot = np.asarray([[c, -s, 0],
                          [s, c, 0],
                          [0, 0, 1]])
    else:
        raise ValueError('axis must be 0, 1, or 2')

    return rot


def affine_from_translation(translation: Any):
    """
        Create an affine translation.

        Parameters
        ----------
        translation: array-like vector of x, y, and z translations

        Returns
        -------
        (4, 4) numpy.ndarray affine matrix

    """
    return affine_from_transform_translation(translation=translation)


def affine_from_transform(transform: Any):
    """Create affine transformation.

        Parameters
        ----------
        transformation: (3, 3) row major array-like transformation

        Returns
        -------
        (4, 4) numpy.ndarray affine matrix

    """
    return affine_from_transform_translation(transform=transform)



"""An executable for finding close-fit boundaries between cortical layer 
polygons.
"""
import logging
import copy as cp
from functools import partial

import shapely

from argschema.argschema_parser import ArgSchemaParser

from neuron_morphology.snap_polygons._schemas import (
    InputParameters, OutputParameters)
from neuron_morphology.snap_polygons._from_lims import FromLimsSource
from neuron_morphology.snap_polygons.geometries import (
    Geometries, find_vertical_surfaces, select_largest_subpolygon
)
from neuron_morphology.snap_polygons.cortex_surfaces import trim_to_close
from neuron_morphology.snap_polygons.image_outputter import ImageOutputter
from neuron_morphology.transforms.geometry import get_vertices_from_two_lines


def run_snap_polygons(
        layer_polygons, 
        pia_surface, 
        wm_surface, 
        layer_order,
        working_scale: float,
        surface_distance_threshold: float,
        multipolygon_error_threshold: float,
        images=None
):
    """Finds and returns close fit boundaries. May write diagnostic images as 
    a side effect.
    """

    # setup input geometries
    geometries = Geometries()
    geometries.register_polygons(layer_polygons)
    
    # setup cortex boundaries
    hull = geometries.convex_hull()
    pia = trim_to_close(hull, surface_distance_threshold, pia_surface["path"])
    white_matter = trim_to_close(
        hull, surface_distance_threshold, wm_surface["path"]
    )
    
    geometries.register_surface("pia", pia)
    geometries.register_surface("wm", white_matter)

    pia_wm_vertices = get_vertices_from_two_lines(
        pia.coords[:], white_matter.coords[:]
    )
    bounds = shapely.geometry.polygon.Polygon(pia_wm_vertices)

    multipolygon_resolver = partial(
        select_largest_subpolygon, 
        error_threshold=multipolygon_error_threshold
    )

    # go!
    result_geos = (
        geometries
        .fill_gaps(working_scale, multipolygon_resolver=multipolygon_resolver)
        .cut(bounds, multipolygon_resolver=multipolygon_resolver)
    )

    # get output surfaces
    boundaries = find_vertical_surfaces(
        result_geos.polygons, 
        layer_order, 
        pia=geometries.surfaces["pia"], 
        white_matter=geometries.surfaces["wm"]
    )
    result_geos.register_surfaces(boundaries)        

    # write results
    outputter = ImageOutputter(
        geometries, result_geos, images
    )
    results = result_geos.to_json()
    results["images"] = outputter.write_images()
    return results


def main():
    """CLI entrypoint for snapping polygons
    """

    class Parser(ArgSchemaParser):
        """An ArgschemaParser that can pull data from LIMS
        """
        default_configurable_sources = \
            ArgSchemaParser.default_configurable_sources + [FromLimsSource]

    parser = Parser(
        schema_type=InputParameters,
        output_schema_type=OutputParameters
    )

    args = cp.deepcopy(parser.args)
    logging.getLogger().setLevel(args.pop("log_level"))

    output = run_snap_polygons(**args)
    output.update({"inputs": parser.args})

    parser.output(output)


if __name__ == "__main__":
    main()



""" A collection of miscellaneous tools used by the feature extractor
"""

from typing import Dict, Any


def unnest(inputs: Dict[str, Any], _prefix="") -> Dict[str, Any]:
    """ Convert nested dictionaries (with string keys) to a dot-notation flat 
    dictionary.

    Parameters
    ---------
    inputs: The dictionary to unnest. Must have all string keys
    _prefix : Used during recursion to build up a dot-notation prefix. Don't 
        argue this yourself!
    
    Returns
    -------
    a flattened dictionary

    """

    unnested = {}
    for key, value in inputs.items():
        if isinstance(key, str):
            if isinstance(value, dict):
                unnested.update(unnest(value, _prefix=f"{key}."))
            else:
                unnested[f"{_prefix}{key}"] = value
        else:
            raise ValueError(f"found non-string key: {key}")
    return unnested



from setuptools import setup, find_packages
from distutils.cmd import Command
import glob
import os


version_path = os.path.join(
    os.path.dirname(__file__), "neuron_morphology", "VERSION.txt"
)
with open(version_path, "r") as version_file:
    version = version_file.read().strip()


with open("requirements.txt", "r") as f:
    required = f.read().splitlines()

readme_path = os.path.join(os.path.dirname(__file__), "README.md")
with open(readme_path, "r") as readme_file:
    readme = readme_file.read()


class CheckVersionCommand(Command):
    description = "Check that this package's version matches a user-supplied version"
    user_options = [
        ("expected-version=", "e", "Compare package version against this value")
    ]

    def initialize_options(self):
        self.package_version = version
        self.expected_version = None

    def finalize_options(self):
        assert self.expected_version is not None
        if self.expected_version[0] == "v":
            self.expected_version = self.expected_version[1:]

    def run(self):
        if self.expected_version != self.package_version:
            raise ValueError(
                f"expected version {self.expected_version}, but this package "
                f"has version {self.package_version}"
            )


setup(
    version=version,
    name="neuron-morphology",
    author="Allen Institute for Brain Science",
    author_email="Marmot@AllenInstitute.onmicrosoft.com",
    packages=find_packages(),
    include_package_data=True,
    description="Tools for working with single-neuron morphological reconstructions",
    long_description=readme,
    extras_require={"streamlines": ["fenics-dolfinx", "python-gmsh", "rasterio"]},
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": [
            "feature_extractor       = neuron_morphology.feature_extractor.__main__:main",
            "layered_point_depths    = neuron_morphology.layered_point_depths.__main__:main",
            "snap_polygons           = neuron_morphology.snap_polygons.__main__:main",
            "apply_affine_transform  = neuron_morphology.transforms.affine_transformer.apply_affine_transform:main",
            "pia_wm_streamlines      = neuron_morphology.transforms.pia_wm_streamlines.calculate_pia_wm_streamlines:main",
            "upright_angle           = neuron_morphology.transforms.upright_angle.compute_angle:main",
            "validate_reconstruction = neuron_morphology.validation.validate_reconstruction:main",
        ]
    },
    keywords=["neuroscience", "bioinformatics", "scientific"],
    install_requires=required,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",  # Allen Institute Software License
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    cmdclass={"check_version": CheckVersionCommand},
)



import pandas as pd
from neuron_morphology.morphology import Morphology
from cloudfiles import CloudFiles
import io
import os
import csv 

SWC_COLUMNS = (
    "id",
    "type",
    "x",
    "y",
    "z",
    "radius",
    "parent",
)
COLUMN_CASTS = {"id": int, "parent": int, "type": int}


def read_swc(path, columns=SWC_COLUMNS, sep=" ", casts=COLUMN_CASTS):

    """Read an swc file into a pandas dataframe"""
    if os.path.dirname(path) == "":
        path = "./" + path

    if "://" not in path:
        path = "file://" + path

    cloudpath, file = os.path.split(path)
    cf = CloudFiles(cloudpath)
    path = io.BytesIO(cf.get(file))

    df = pd.read_csv(path, names=columns, comment="#", sep=sep, index_col=False)
    apply_casts(df, casts)
    return df


def write_swc(
    data, path, comments=None, sep=" ", columns=SWC_COLUMNS, casts=COLUMN_CASTS
):

    """Write an swc file from a pandas dataframe"""

    if comments is None:
        comments = []
    comments = ["# " + comment + "\n" for comment in comments]

    apply_casts(data, casts)

    data = data[[col for col in columns]]
    if os.path.dirname(path) == "":
        path = "./" + path
    if "://" not in path:
        path = "file://" + path
    cloudpath, file = os.path.split(path)
    cf = CloudFiles(cloudpath)
    buffer = io.BytesIO()
    charset = "utf-8"
    wrapper = io.TextIOWrapper(buffer, encoding=charset,newline='',
                               line_buffering=True)
    wrapper.writelines(comments)
    data.to_csv(wrapper, sep=sep, index=False, header=None, mode="a")
    wrapper.flush()
    buffer.seek(0)
    cf.put(file, buffer.getvalue(), content_type="application/x-swc")


def apply_casts(df, casts):

    for key, typ in casts.items():
        df[key] = df[key].astype(typ)


def morphology_from_swc(swc_path):

    swc_data = read_swc(swc_path, sep=" ")

    nodes = swc_data.to_dict("records")
    for node in nodes:
        # unfortunately, pandas automatically promotes numeric types to float in to_dict
        node["parent"] = int(node["parent"])
        node["id"] = int(node["id"])
        node["type"] = int(node["type"])

    return Morphology(
        nodes,
        node_id_cb=lambda node: node["id"],
        parent_id_cb=lambda node: node["parent"],
    )


def morphology_to_swc(morphology, swc_path, comments=None):
    """
    Write an swc file from a morphology object
    """

    df = pd.DataFrame(morphology.nodes())
    write_swc(df, swc_path, comments=comments)



# -*- coding: utf-8 -*-

"""Top-level package for neuron_morphology."""

import os
from logging.config import fileConfig


version_path = os.path.join(os.path.dirname(__file__), "VERSION.txt")
with open(version_path, "r") as version_file:
    __version__ = version_file.read().strip()


fileConfig(os.path.join(
    os.path.dirname(__file__), 
    'logging_config.ini')
)



import csv
import neuron_morphology.constants as constants
from neuron_morphology.validation.result import InvalidMarkerFile
from neuron_morphology.validation.result import MarkerValidationError


class Marker(dict):
    """ Simple dictionary class for handling reconstruction marker objects. """

    def __init__(self, *args, **kwargs):
        super(Marker, self).__init__(*args, **kwargs)

        self['original_x'] = self['x']
        self['original_y'] = self['y']
        self['original_z'] = self['z']

        # marker file x,y,z coordinates are offset by a single image-space pixel
        self['x'] -= constants.SPACING[0]
        self['y'] -= constants.SPACING[1]
        self['z'] -= constants.SPACING[2]


def read_marker_file(file_name):
    """ read in a marker file and return a list of dictionaries """

    with open(file_name, 'r') as f:
        rows = csv.DictReader((r for r in f if not r.startswith('#')), fieldnames=['x', 'y', 'z', 'radius', 'shape',
                                                                                   'name', 'comment', 'color_r',
                                                                                   'color_g', 'color_b'])

        markers = []

        for r in rows:
            try:
                markers.append(Marker({'x': float(r['x']),
                                       'y': float(r['y']),
                                       'z': float(r['z']),
                                       'name': int(r['name'])}))

            except ValueError:
                message = "Failed to parse row. One of (x, y, z, name) is missing or invalid"
                raise InvalidMarkerFile([MarkerValidationError(message, r, "Error")])

        return markers
